"""
MetaFormer, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/metaformer.py
and
https://github.com/sail-sg/metaformer

Paper "MetaFormer Baselines for Vision", https://arxiv.org/abs/2210.13452
"""

# Reference license: Apache-2.0 (both)

from collections import OrderedDict
from collections.abc import Callable
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.layers import LayerNorm2d
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType


class Downsample(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
        norm_layer: Optional[Callable[..., nn.Module]],
    ) -> None:
        super().__init__()
        self.norm = norm_layer(in_channels) if norm_layer else nn.Identity()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=kernel_size, stride=stride, padding=padding)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.norm(x)
        x = self.conv(x)
        return x


class ConvMLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: int,
        out_features: int,
        act_layer: Callable[..., nn.Module],
        bias: bool,
        dropout: float,
    ) -> None:
        super().__init__()
        self.fc1 = nn.Conv2d(in_features, hidden_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=bias)
        self.act = act_layer()
        self.drop1 = nn.Dropout(dropout)
        self.fc2 = nn.Conv2d(
            hidden_features, out_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=bias
        )
        self.drop2 = nn.Dropout(dropout)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop1(x)
        x = self.fc2(x)
        x = self.drop2(x)

        return x


class Scale(nn.Module):
    def __init__(self, dim: int, init_value: float) -> None:
        super().__init__()
        self.scale = nn.Parameter(init_value * torch.ones(dim, 1, 1), requires_grad=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * self.scale


class SquaredReLU(nn.Module):
    """
    Squared ReLU: https://arxiv.org/abs/2109.08668
    """

    def __init__(self, inplace: bool = False) -> None:
        super().__init__()
        self.relu = nn.ReLU(inplace=inplace)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.square(self.relu(x))


class StarReLU(nn.Module):
    def __init__(self, scale_value: float = 1.0, bias_value: float = 0.0, inplace: bool = False) -> None:
        super().__init__()
        self.relu = nn.ReLU(inplace=inplace)
        self.scale = nn.Parameter(scale_value * torch.ones(1), requires_grad=True)
        self.bias = nn.Parameter(bias_value * torch.ones(1), requires_grad=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.scale * self.relu(x) ** 2 + self.bias


class Attention(nn.Module):
    def __init__(self, dim: int, head_dim: int = 32, attn_drop: float = 0.0, proj_drop: float = 0.0) -> None:
        super().__init__()

        self.head_dim = head_dim
        self.scale = head_dim**-0.5
        self.num_heads = dim // head_dim
        self.attention_dim = self.num_heads * self.head_dim

        self.qkv = nn.Linear(dim, self.attention_dim * 3, bias=False)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(self.attention_dim, dim, bias=False)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.permute(0, 2, 3, 1)  # (N, C, H, W) -> (N, H, W, C)
        (B, H, W, _) = x.shape
        N = H * W
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, self.head_dim).permute(2, 0, 3, 1, 4)
        (q, k, v) = qkv.unbind(0)

        x = F.scaled_dot_product_attention(  # pylint: disable=not-callable
            q, k, v, dropout_p=self.attn_drop.p if self.training else 0.0, scale=self.scale
        )

        x = x.transpose(1, 2).reshape(B, H, W, self.attention_dim)
        x = self.proj(x)
        x = self.proj_drop(x)
        x = x.permute(0, 3, 1, 2)  # (N, H, W, C) -> (N, C, H, W)

        return x


class GroupNorm1(nn.GroupNorm):
    def __init__(self, num_channels: int) -> None:
        super().__init__(num_groups=1, num_channels=num_channels, eps=1e-6)

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # pylint: disable=arguments-renamed
        return F.group_norm(x, self.num_groups, self.weight, self.bias, self.eps)


class GroupNorm1NoBias(nn.GroupNorm):
    def __init__(self, num_channels: int) -> None:
        super().__init__(num_groups=1, num_channels=num_channels, eps=1e-6)
        self.bias = None

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # pylint: disable=arguments-renamed
        return F.group_norm(x, self.num_groups, self.weight, self.bias, self.eps)


class LayerNorm2dNoBias(LayerNorm2d):
    def __init__(self, num_channels: int) -> None:
        super().__init__(num_channels, eps=1e-6)
        self.bias = None


class SepConv(nn.Module):
    """
    Same as MobileNet v2 inverted separable convolution without the normalization
    """

    def __init__(
        self,
        dim: int,
        expansion_ratio: float = 2,
        act1_layer: Callable[..., nn.Module] = StarReLU,
        kernel_size: tuple[int, int] = (7, 7),
        padding: tuple[int, int] = (3, 3),
        **kwargs: Any,
    ) -> None:
        super().__init__()
        mid_channels = int(expansion_ratio * dim)
        self.pw_conv1 = nn.Conv2d(dim, mid_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False)
        self.act = act1_layer()
        self.dwconv = nn.Conv2d(
            mid_channels,
            mid_channels,
            kernel_size=kernel_size,
            stride=(1, 1),
            padding=padding,
            groups=mid_channels,
            bias=False,
        )
        self.pw_conv2 = nn.Conv2d(mid_channels, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.pw_conv1(x)
        x = self.act(x)
        x = self.dwconv(x)
        x = self.pw_conv2(x)

        return x


class Pooling(nn.Module):
    """
    PoolFormer: https://arxiv.org/abs/2111.11418
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__()
        self.pool = nn.AvgPool2d(kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), count_include_pad=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.pool(x) - x


class MetaFormerBlock(nn.Module):
    """
    Implementation of one MetaFormer block.
    """

    def __init__(
        self,
        dim: int,
        token_mixer: Callable[..., nn.Module],
        mlp_act: Callable[..., nn.Module],
        mlp_bias: bool,
        norm_layer: Callable[..., nn.Module],
        proj_drop: float,
        drop_path: float,
        layer_scale_init_value: Optional[float],
        res_scale_init_value: Optional[float],
    ) -> None:
        super().__init__()

        self.norm1 = norm_layer(dim)
        self.token_mixer = token_mixer(dim=dim, proj_drop=proj_drop)
        self.drop_path1 = StochasticDepth(drop_path, mode="row")
        if layer_scale_init_value is None:
            self.layer_scale1 = nn.Identity()
            self.layer_scale2 = nn.Identity()
        else:
            self.layer_scale1 = Scale(dim=dim, init_value=layer_scale_init_value)
            self.layer_scale2 = Scale(dim=dim, init_value=layer_scale_init_value)

        if res_scale_init_value is None:
            self.res_scale1 = nn.Identity()
            self.res_scale2 = nn.Identity()
        else:
            self.res_scale1 = Scale(dim=dim, init_value=res_scale_init_value)
            self.res_scale2 = Scale(dim=dim, init_value=res_scale_init_value)

        self.norm2 = norm_layer(dim)
        self.mlp = ConvMLP(dim, 4 * dim, dim, act_layer=mlp_act, bias=mlp_bias, dropout=proj_drop)
        self.drop_path2 = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.res_scale1(x) + self.layer_scale1(self.drop_path1(self.token_mixer(self.norm1(x))))
        x = self.res_scale2(x) + self.layer_scale2(self.drop_path2(self.mlp(self.norm2(x))))

        return x


class MetaFormerStage(nn.Module):
    def __init__(
        self,
        in_chs: int,
        out_chs: int,
        depth: int,
        token_mixer: Callable[..., nn.Module],
        mlp_act: Callable[..., nn.Module],
        mlp_bias: bool,
        downsample_norm: Callable[..., nn.Module],
        norm_layer: Callable[..., nn.Module],
        proj_drop: float,
        dp_rates: list[float],
        layer_scale_init_value: Optional[float],
        res_scale_init_value: Optional[float],
    ) -> None:
        super().__init__()

        if in_chs == out_chs:
            self.downsample = nn.Identity()
        else:
            self.downsample = Downsample(
                in_chs,
                out_chs,
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                norm_layer=downsample_norm,
            )

        layers = []
        for i in range(depth):
            layers.append(
                MetaFormerBlock(
                    dim=out_chs,
                    token_mixer=token_mixer,
                    mlp_act=mlp_act,
                    mlp_bias=mlp_bias,
                    norm_layer=norm_layer,
                    proj_drop=proj_drop,
                    drop_path=dp_rates[i],
                    layer_scale_init_value=layer_scale_init_value,
                    res_scale_init_value=res_scale_init_value,
                )
            )

        self.blocks = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)

        return x


class MetaFormer(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.stage(\d+)\.blocks\.(\d+)"

    # pylint: disable=too-many-locals,too-many-branches
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        depths: list[int] = self.config["depths"]
        dims: list[int] = self.config["dims"]
        token_mixer_names: list[str] = self.config["token_mixer_names"]
        mlp_act_name: str = self.config["mlp_act_name"]
        mlp_bias: bool = self.config["mlp_bias"]
        layer_scale_init_values: list[Optional[float]] = self.config["layer_scale_init_values"]
        res_scale_init_values: list[Optional[float]] = self.config["res_scale_init_values"]
        norm_layer_names: list[str] = self.config["norm_layer_names"]
        downsample_norm_name: str = self.config["downsample_norm_name"]
        drop_path_rate: float = self.config["drop_path_rate"]
        use_mlp_head: bool = self.config["use_mlp_head"]
        mlp_head_dropout: float = self.config["mlp_head_dropout"]

        token_mixers: list[type[nn.Module]] = []
        for token_mixer_name in token_mixer_names:
            if token_mixer_name == "Pooling":  # nosec
                token_mixers.append(Pooling)
            elif token_mixer_name == "SepConv":  # nosec
                token_mixers.append(SepConv)
            elif token_mixer_name == "Attention":  # nosec
                token_mixers.append(Attention)
            else:
                raise ValueError(f"Unknown token_mixer_name '{token_mixer_name}'")

        if mlp_act_name == "StarReLU":
            mlp_act = StarReLU
        elif mlp_act_name == "GELU":
            mlp_act = nn.GELU
        else:
            raise ValueError(f"Unknown mlp_act_name '{mlp_act_name}'")

        norm_layers: list[type[nn.Module]] = []
        for norm_layer_name in norm_layer_names:
            if norm_layer_name == "GroupNorm1":
                norm_layers.append(GroupNorm1)
            elif norm_layer_name == "GroupNorm1NoBias":
                norm_layers.append(GroupNorm1NoBias)
            elif norm_layer_name == "LayerNorm2dNoBias":
                norm_layers.append(LayerNorm2dNoBias)
            else:
                raise ValueError(f"Unknown norm_layer_name '{norm_layer_name}'")

        if downsample_norm_name == "Identity":
            downsample_norm = nn.Identity
        elif downsample_norm_name == "LayerNorm2dNoBias":
            downsample_norm = LayerNorm2dNoBias
        else:
            raise ValueError(f"Unknown downsample_norm_name '{downsample_norm_name}'")

        self.use_mlp_head = use_mlp_head
        self.mlp_head_dropout = mlp_head_dropout
        self.stem = nn.Sequential(
            nn.Conv2d(self.input_channels, dims[0], kernel_size=(7, 7), stride=(4, 4), padding=(2, 2)),
            downsample_norm(dims[0]),
        )

        num_stages = len(depths)
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        prev_dim = dims[0]
        dp_rates = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]
        for i in range(num_stages):
            stages[f"stage{i+1}"] = MetaFormerStage(
                prev_dim,
                dims[i],
                depth=depths[i],
                token_mixer=token_mixers[i],
                mlp_act=mlp_act,
                mlp_bias=mlp_bias,
                proj_drop=0.0,
                dp_rates=dp_rates[i],
                layer_scale_init_value=layer_scale_init_values[i],
                res_scale_init_value=res_scale_init_values[i],
                downsample_norm=downsample_norm,
                norm_layer=norm_layers[i],
            )
            return_channels.append(dims[i])
            prev_dim = dims[i]

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            LayerNorm2d(dims[-1], eps=1e-6),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = dims[-1]
        self.classifier = self.create_classifier()

        self.stem_stride = 4
        self.stem_width = dims[0]
        self.encoding_size = dims[-1]

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.trunc_normal_(m.weight, std=0.02)

            elif isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(x, mask, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token)
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def create_classifier(self, embed_dim: Optional[int] = None) -> nn.Module:
        if self.num_classes == 0:
            return nn.Identity()

        if embed_dim is None:
            embed_dim = self.embedding_size

        if self.use_mlp_head is False:
            return nn.Linear(embed_dim, self.num_classes)

        return nn.Sequential(
            nn.Dropout(self.mlp_head_dropout),
            nn.Linear(embed_dim, 4 * embed_dim),
            SquaredReLU(),
            nn.LayerNorm(4 * embed_dim),
            nn.Linear(4 * embed_dim, self.num_classes),
        )


# PoolFormer v1
registry.register_model_config(
    "poolformer_v1_s12",
    MetaFormer,
    config={
        "depths": [2, 2, 6, 2],
        "dims": [64, 128, 320, 512],
        "token_mixer_names": ["Pooling", "Pooling", "Pooling", "Pooling"],
        "mlp_act_name": "GELU",
        "mlp_bias": True,
        "layer_scale_init_values": [1e-5, 1e-5, 1e-5, 1e-5],
        "res_scale_init_values": [None, None, None, None],
        "norm_layer_names": ["GroupNorm1", "GroupNorm1", "GroupNorm1", "GroupNorm1"],
        "downsample_norm_name": "Identity",
        "drop_path_rate": 0.1,
        "use_mlp_head": False,
        "mlp_head_dropout": 0.0,
    },
)
registry.register_model_config(
    "poolformer_v1_s24",
    MetaFormer,
    config={
        "depths": [4, 4, 12, 4],
        "dims": [64, 128, 320, 512],
        "token_mixer_names": ["Pooling", "Pooling", "Pooling", "Pooling"],
        "mlp_act_name": "GELU",
        "mlp_bias": True,
        "layer_scale_init_values": [1e-5, 1e-5, 1e-5, 1e-5],
        "res_scale_init_values": [None, None, None, None],
        "norm_layer_names": ["GroupNorm1", "GroupNorm1", "GroupNorm1", "GroupNorm1"],
        "downsample_norm_name": "Identity",
        "drop_path_rate": 0.1,
        "use_mlp_head": False,
        "mlp_head_dropout": 0.0,
    },
)
registry.register_model_config(
    "poolformer_v1_s36",
    MetaFormer,
    config={
        "depths": [6, 6, 18, 6],
        "dims": [64, 128, 320, 512],
        "token_mixer_names": ["Pooling", "Pooling", "Pooling", "Pooling"],
        "mlp_act_name": "GELU",
        "mlp_bias": True,
        "layer_scale_init_values": [1e-6, 1e-6, 1e-6, 1e-6],
        "res_scale_init_values": [None, None, None, None],
        "norm_layer_names": ["GroupNorm1", "GroupNorm1", "GroupNorm1", "GroupNorm1"],
        "downsample_norm_name": "Identity",
        "drop_path_rate": 0.2,
        "use_mlp_head": False,
        "mlp_head_dropout": 0.0,
    },
)
registry.register_model_config(
    "poolformer_v1_m36",
    MetaFormer,
    config={
        "depths": [6, 6, 18, 6],
        "dims": [96, 192, 384, 768],
        "token_mixer_names": ["Pooling", "Pooling", "Pooling", "Pooling"],
        "mlp_act_name": "GELU",
        "mlp_bias": True,
        "layer_scale_init_values": [1e-6, 1e-6, 1e-6, 1e-6],
        "res_scale_init_values": [None, None, None, None],
        "norm_layer_names": ["GroupNorm1", "GroupNorm1", "GroupNorm1", "GroupNorm1"],
        "downsample_norm_name": "Identity",
        "drop_path_rate": 0.3,
        "use_mlp_head": False,
        "mlp_head_dropout": 0.0,
    },
)
registry.register_model_config(
    "poolformer_v1_m48",
    MetaFormer,
    config={
        "depths": [8, 8, 24, 8],
        "dims": [96, 192, 384, 768],
        "token_mixer_names": ["Pooling", "Pooling", "Pooling", "Pooling"],
        "mlp_act_name": "GELU",
        "mlp_bias": True,
        "layer_scale_init_values": [1e-6, 1e-6, 1e-6, 1e-6],
        "res_scale_init_values": [None, None, None, None],
        "norm_layer_names": ["GroupNorm1", "GroupNorm1", "GroupNorm1", "GroupNorm1"],
        "downsample_norm_name": "Identity",
        "drop_path_rate": 0.4,
        "use_mlp_head": False,
        "mlp_head_dropout": 0.0,
    },
)

# PoolFormer v2
registry.register_model_config(
    "poolformer_v2_s12",
    MetaFormer,
    config={
        "depths": [2, 2, 6, 2],
        "dims": [64, 128, 320, 512],
        "token_mixer_names": ["Pooling", "Pooling", "Pooling", "Pooling"],
        "mlp_act_name": "StarReLU",
        "mlp_bias": False,
        "layer_scale_init_values": [None, None, None, None],
        "res_scale_init_values": [None, None, 1.0, 1.0],
        "norm_layer_names": ["GroupNorm1NoBias", "GroupNorm1NoBias", "GroupNorm1NoBias", "GroupNorm1NoBias"],
        "downsample_norm_name": "LayerNorm2dNoBias",
        "drop_path_rate": 0.1,
        "use_mlp_head": False,
        "mlp_head_dropout": 0.0,
    },
)
registry.register_model_config(
    "poolformer_v2_s24",
    MetaFormer,
    config={
        "depths": [4, 4, 12, 4],
        "dims": [64, 128, 320, 512],
        "token_mixer_names": ["Pooling", "Pooling", "Pooling", "Pooling"],
        "mlp_act_name": "StarReLU",
        "mlp_bias": False,
        "layer_scale_init_values": [None, None, None, None],
        "res_scale_init_values": [None, None, 1.0, 1.0],
        "norm_layer_names": ["GroupNorm1NoBias", "GroupNorm1NoBias", "GroupNorm1NoBias", "GroupNorm1NoBias"],
        "downsample_norm_name": "LayerNorm2dNoBias",
        "drop_path_rate": 0.1,
        "use_mlp_head": False,
        "mlp_head_dropout": 0.0,
    },
)
registry.register_model_config(
    "poolformer_v2_s36",
    MetaFormer,
    config={
        "depths": [6, 6, 18, 6],
        "dims": [64, 128, 320, 512],
        "token_mixer_names": ["Pooling", "Pooling", "Pooling", "Pooling"],
        "mlp_act_name": "StarReLU",
        "mlp_bias": False,
        "layer_scale_init_values": [None, None, None, None],
        "res_scale_init_values": [None, None, 1.0, 1.0],
        "norm_layer_names": ["GroupNorm1NoBias", "GroupNorm1NoBias", "GroupNorm1NoBias", "GroupNorm1NoBias"],
        "downsample_norm_name": "LayerNorm2dNoBias",
        "drop_path_rate": 0.2,
        "use_mlp_head": False,
        "mlp_head_dropout": 0.0,
    },
)
registry.register_model_config(
    "poolformer_v2_m36",
    MetaFormer,
    config={
        "depths": [6, 6, 18, 6],
        "dims": [96, 192, 384, 768],
        "token_mixer_names": ["Pooling", "Pooling", "Pooling", "Pooling"],
        "mlp_act_name": "StarReLU",
        "mlp_bias": False,
        "layer_scale_init_values": [None, None, None, None],
        "res_scale_init_values": [None, None, 1.0, 1.0],
        "norm_layer_names": ["GroupNorm1NoBias", "GroupNorm1NoBias", "GroupNorm1NoBias", "GroupNorm1NoBias"],
        "downsample_norm_name": "LayerNorm2dNoBias",
        "drop_path_rate": 0.3,
        "use_mlp_head": False,
        "mlp_head_dropout": 0.0,
    },
)
registry.register_model_config(
    "poolformer_v2_m48",
    MetaFormer,
    config={
        "depths": [8, 8, 24, 8],
        "dims": [96, 192, 384, 768],
        "token_mixer_names": ["Pooling", "Pooling", "Pooling", "Pooling"],
        "mlp_act_name": "StarReLU",
        "mlp_bias": False,
        "layer_scale_init_values": [None, None, None, None],
        "res_scale_init_values": [None, None, 1.0, 1.0],
        "norm_layer_names": ["GroupNorm1NoBias", "GroupNorm1NoBias", "GroupNorm1NoBias", "GroupNorm1NoBias"],
        "downsample_norm_name": "LayerNorm2dNoBias",
        "drop_path_rate": 0.4,
        "use_mlp_head": False,
        "mlp_head_dropout": 0.0,
    },
)

# ConvFormer
registry.register_model_config(
    "convformer_s18",
    MetaFormer,
    config={
        "depths": [3, 3, 9, 3],
        "dims": [64, 128, 320, 512],
        "token_mixer_names": ["SepConv", "SepConv", "SepConv", "SepConv"],
        "mlp_act_name": "StarReLU",
        "mlp_bias": False,
        "layer_scale_init_values": [None, None, None, None],
        "res_scale_init_values": [None, None, 1.0, 1.0],
        "norm_layer_names": ["LayerNorm2dNoBias", "LayerNorm2dNoBias", "LayerNorm2dNoBias", "LayerNorm2dNoBias"],
        "downsample_norm_name": "LayerNorm2dNoBias",
        "drop_path_rate": 0.2,
        "use_mlp_head": True,
        "mlp_head_dropout": 0.0,
    },
)
registry.register_model_config(
    "convformer_s36",
    MetaFormer,
    config={
        "depths": [3, 12, 18, 3],
        "dims": [64, 128, 320, 512],
        "token_mixer_names": ["SepConv", "SepConv", "SepConv", "SepConv"],
        "mlp_act_name": "StarReLU",
        "mlp_bias": False,
        "layer_scale_init_values": [None, None, None, None],
        "res_scale_init_values": [None, None, 1.0, 1.0],
        "norm_layer_names": ["LayerNorm2dNoBias", "LayerNorm2dNoBias", "LayerNorm2dNoBias", "LayerNorm2dNoBias"],
        "downsample_norm_name": "LayerNorm2dNoBias",
        "drop_path_rate": 0.3,
        "use_mlp_head": True,
        "mlp_head_dropout": 0.0,
    },
)
registry.register_model_config(
    "convformer_m36",
    MetaFormer,
    config={
        "depths": [3, 12, 18, 3],
        "dims": [96, 192, 384, 576],
        "token_mixer_names": ["SepConv", "SepConv", "SepConv", "SepConv"],
        "mlp_act_name": "StarReLU",
        "mlp_bias": False,
        "layer_scale_init_values": [None, None, None, None],
        "res_scale_init_values": [None, None, 1.0, 1.0],
        "norm_layer_names": ["LayerNorm2dNoBias", "LayerNorm2dNoBias", "LayerNorm2dNoBias", "LayerNorm2dNoBias"],
        "downsample_norm_name": "LayerNorm2dNoBias",
        "drop_path_rate": 0.4,
        "use_mlp_head": True,
        "mlp_head_dropout": 0.0,
    },
)
registry.register_model_config(
    "convformer_b36",
    MetaFormer,
    config={
        "depths": [3, 12, 18, 3],
        "dims": [128, 256, 512, 768],
        "token_mixer_names": ["SepConv", "SepConv", "SepConv", "SepConv"],
        "mlp_act_name": "StarReLU",
        "mlp_bias": False,
        "layer_scale_init_values": [None, None, None, None],
        "res_scale_init_values": [None, None, 1.0, 1.0],
        "norm_layer_names": ["LayerNorm2dNoBias", "LayerNorm2dNoBias", "LayerNorm2dNoBias", "LayerNorm2dNoBias"],
        "downsample_norm_name": "LayerNorm2dNoBias",
        "drop_path_rate": 0.6,
        "use_mlp_head": True,
        "mlp_head_dropout": 0.0,
    },
)

# CAFormer
registry.register_model_config(
    "caformer_s18",
    MetaFormer,
    config={
        "depths": [3, 3, 9, 3],
        "dims": [64, 128, 320, 512],
        "token_mixer_names": ["SepConv", "SepConv", "Attention", "Attention"],
        "mlp_act_name": "StarReLU",
        "mlp_bias": False,
        "layer_scale_init_values": [None, None, None, None],
        "res_scale_init_values": [None, None, 1.0, 1.0],
        "norm_layer_names": ["LayerNorm2dNoBias", "LayerNorm2dNoBias", "LayerNorm2dNoBias", "LayerNorm2dNoBias"],
        "downsample_norm_name": "LayerNorm2dNoBias",
        "drop_path_rate": 0.15,
        "use_mlp_head": True,
        "mlp_head_dropout": 0.0,
    },
)
registry.register_model_config(
    "caformer_s36",
    MetaFormer,
    config={
        "depths": [3, 12, 18, 3],
        "dims": [64, 128, 320, 512],
        "token_mixer_names": ["SepConv", "SepConv", "Attention", "Attention"],
        "mlp_act_name": "StarReLU",
        "mlp_bias": False,
        "layer_scale_init_values": [None, None, None, None],
        "res_scale_init_values": [None, None, 1.0, 1.0],
        "norm_layer_names": ["LayerNorm2dNoBias", "LayerNorm2dNoBias", "LayerNorm2dNoBias", "LayerNorm2dNoBias"],
        "downsample_norm_name": "LayerNorm2dNoBias",
        "drop_path_rate": 0.3,
        "use_mlp_head": True,
        "mlp_head_dropout": 0.4,
    },
)
registry.register_model_config(
    "caformer_m36",
    MetaFormer,
    config={
        "depths": [3, 12, 18, 3],
        "dims": [96, 192, 384, 576],
        "token_mixer_names": ["SepConv", "SepConv", "Attention", "Attention"],
        "mlp_act_name": "StarReLU",
        "mlp_bias": False,
        "layer_scale_init_values": [None, None, None, None],
        "res_scale_init_values": [None, None, 1.0, 1.0],
        "norm_layer_names": ["LayerNorm2dNoBias", "LayerNorm2dNoBias", "LayerNorm2dNoBias", "LayerNorm2dNoBias"],
        "downsample_norm_name": "LayerNorm2dNoBias",
        "drop_path_rate": 0.4,
        "use_mlp_head": True,
        "mlp_head_dropout": 0.4,
    },
)
registry.register_model_config(
    "caformer_b36",
    MetaFormer,
    config={
        "depths": [3, 12, 18, 3],
        "dims": [128, 256, 512, 768],
        "token_mixer_names": ["SepConv", "SepConv", "Attention", "Attention"],
        "mlp_act_name": "StarReLU",
        "mlp_bias": False,
        "layer_scale_init_values": [None, None, None, None],
        "res_scale_init_values": [None, None, 1.0, 1.0],
        "norm_layer_names": ["LayerNorm2dNoBias", "LayerNorm2dNoBias", "LayerNorm2dNoBias", "LayerNorm2dNoBias"],
        "downsample_norm_name": "LayerNorm2dNoBias",
        "drop_path_rate": 0.6,
        "use_mlp_head": True,
        "mlp_head_dropout": 0.5,
    },
)

registry.register_weights(
    "poolformer_v1_s12_il-common",
    {
        "description": "PoolFormer v1 small 12 layers trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 44.3,
                "sha256": "4a968765a2a50957c0cdb24c9c57979e24b408161a915552dcd1b3650721b831",
            }
        },
        "net": {"network": "poolformer_v1_s12", "tag": "il-common"},
    },
)
registry.register_weights(
    "poolformer_v2_s12_il-common",
    {
        "description": "PoolFormer v2 small 12 layers trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 44.2,
                "sha256": "13f44e64994336c91e9ae0778c60c32372baf0b3ea86070df36d69cdf5bc7df6",
            }
        },
        "net": {"network": "poolformer_v2_s12", "tag": "il-common"},
    },
)



"""
EfficientViM, adapted from
https://github.com/mlvlab/EfficientViM/blob/main/classification/models/EfficientViM.py

Paper "EfficientViM: Efficient Vision Mamba with Hidden State Mixer based State Space Duality",
https://arxiv.org/abs/2411.15241

Changes from original:
* No distillation heads
"""

# Reference license: MIT

from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import SqueezeExcitation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class LayerNorm1d(nn.Module):
    """
    LayerNorm for channels of 1D tensor (B C L)
    """

    def __init__(self, num_channels: int, eps: float = 1e-5, affine: bool = True) -> None:
        super().__init__()
        self.num_channels = num_channels
        self.eps = eps
        self.affine = affine

        if self.affine is True:
            self.weight = nn.Parameter(torch.ones(1, num_channels, 1))
            self.bias = nn.Parameter(torch.zeros(1, num_channels, 1))
        else:
            self.register_parameter("weight", None)
            self.register_parameter("bias", None)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        mean = x.mean(dim=1, keepdim=True)  # (B, 1, H, W)
        var = x.var(dim=1, keepdim=True, unbiased=False)  # (B, 1, H, W)

        x_normalized = (x - mean) / torch.sqrt(var + self.eps)  # (B, C, H, W)

        if self.affine is True:
            x_normalized = x_normalized * self.weight + self.bias

        return x_normalized


class LayerNorm2d(nn.Module):
    """
    LayerNorm for channels of 2D tensor (B C H W)
    """

    def __init__(self, num_channels: int, eps: float = 1e-5, affine: bool = True) -> None:
        super().__init__()
        self.num_channels = num_channels
        self.eps = eps
        self.affine = affine

        if self.affine is True:
            self.weight = nn.Parameter(torch.ones(1, num_channels, 1, 1))
            self.bias = nn.Parameter(torch.zeros(1, num_channels, 1, 1))
        else:
            self.register_parameter("weight", None)
            self.register_parameter("bias", None)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        mean = x.mean(dim=1, keepdim=True)  # (B, 1, H, W)
        var = x.var(dim=1, keepdim=True, unbiased=False)  # (B, 1, H, W)

        x_normalized = (x - mean) / torch.sqrt(var + self.eps)  # (B, C, H, W)

        if self.affine is True:
            x_normalized = x_normalized * self.weight + self.bias

        return x_normalized


class Conv2dNorm(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
        groups: int = 1,
        zero_bn_init: bool = False,
    ) -> None:
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            groups=groups,
            bias=False,
        )
        self.norm = nn.BatchNorm2d(out_channels)

        if zero_bn_init is True:
            nn.init.zeros_(self.norm.weight)

        nn.init.zeros_(self.norm.bias)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        x = self.norm(x)
        return x


class FFN(nn.Module):
    def __init__(self, in_dim: int, dim: int) -> None:
        super().__init__()
        self.fc1 = Conv2dNorm(in_dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.fc2 = Conv2dNorm(dim, in_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), zero_bn_init=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = F.relu(self.fc1(x))
        x = self.fc2(x)
        return x


class Stem(nn.Module):
    def __init__(self, in_dim: int, dim: int) -> None:
        super().__init__()
        self.conv = nn.Sequential(
            Conv2dNormActivation(in_dim, dim // 8, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
            Conv2dNormActivation(dim // 8, dim // 4, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
            Conv2dNormActivation(dim // 4, dim // 2, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
            Conv2dNormActivation(
                dim // 2, dim, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), activation_layer=None
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.conv(x)


class PatchMerging(nn.Module):
    def __init__(self, in_dim: int, out_dim: int, ratio: float) -> None:
        super().__init__()
        hidden_dim = int(out_dim * ratio)

        self.dwconv1 = Conv2dNorm(in_dim, in_dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=in_dim)
        self.conv = nn.Sequential(
            Conv2dNorm(in_dim, hidden_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)),
            nn.ReLU(),
            Conv2dNorm(hidden_dim, hidden_dim, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), groups=hidden_dim),
            nn.ReLU(),
            SqueezeExcitation(hidden_dim, hidden_dim // 4),
            Conv2dNorm(hidden_dim, out_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)),
        )
        self.dwconv2 = Conv2dNorm(out_dim, out_dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=out_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.dwconv1(x)
        x = self.conv(x)
        x = x + self.dwconv2(x)

        return x


# pylint: disable=invalid-name
class HSMSSD(nn.Module):
    def __init__(self, d_model: int, ssd_expand: float, A_init_range: tuple[int, int], state_dim: int) -> None:
        super().__init__()
        self.d_inner = int(ssd_expand * d_model)
        self.state_dim = state_dim

        conv_dim = state_dim * 3
        self.bc_dt_proj = nn.Conv1d(d_model, 3 * state_dim, kernel_size=1, stride=1, padding=0, bias=False)
        self.dw = nn.Conv2d(
            conv_dim, conv_dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=conv_dim, bias=False
        )
        self.hz_proj = nn.Conv1d(d_model, 2 * self.d_inner, kernel_size=1, stride=1, padding=0, bias=False)
        self.out_proj = nn.Conv1d(self.d_inner, d_model, kernel_size=1, stride=1, padding=0, bias=False)

        A = torch.empty(state_dim, dtype=torch.float32).uniform_(*A_init_range)
        self.A = nn.Parameter(A)
        self.act = nn.SiLU()
        self.D = nn.Parameter(torch.ones(1))
        self.D._no_weight_decay = True  # NOTE: Should this be supported ? is this a thing ?

    def forward(self, x: torch.Tensor, H: int, W: int) -> tuple[torch.Tensor, torch.Tensor]:
        batch_size = x.size(0)

        bc_dt = self.dw(self.bc_dt_proj(x).view(batch_size, -1, H, W)).flatten(2)
        (B, C, dt) = torch.split(bc_dt, [self.state_dim, self.state_dim, self.state_dim], dim=1)
        A = F.softmax(dt + self.A.view(1, -1, 1), dim=-1)

        AB = A * B
        h = x @ AB.transpose(-2, -1)

        (h, z) = torch.split(self.hz_proj(h), [self.d_inner, self.d_inner], dim=1)
        h = self.out_proj(h * self.act(z) + h * self.D)
        y = h @ C

        y = y.view(batch_size, -1, H, W).contiguous()  # + x * self.D  # B C H W

        return (y, h)


class EfficientViMBlock(nn.Module):
    def __init__(
        self, dim: int, mlp_ratio: float, ssd_expand: float, A_init_range: tuple[int, int], state_dim: int
    ) -> None:
        super().__init__()
        self.dim = dim
        self.mlp_ratio = mlp_ratio

        self.mixer = HSMSSD(dim, ssd_expand=ssd_expand, A_init_range=A_init_range, state_dim=state_dim)
        self.norm = LayerNorm1d(dim)

        self.dwconv1 = Conv2dNorm(
            dim, dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=dim, zero_bn_init=True
        )
        self.dwconv2 = Conv2dNorm(
            dim, dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=dim, zero_bn_init=True
        )

        self.ffn = FFN(in_dim=dim, dim=int(dim * mlp_ratio))

        # LayerScale
        self.alpha = nn.Parameter(1e-4 * torch.ones(4, dim))

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        alpha = torch.sigmoid(self.alpha).view(4, -1, 1, 1)

        # DWconv1
        x = (1 - alpha[0]) * x + alpha[0] * self.dwconv1(x)

        # HSM-SSD
        (H, W) = x.shape[-2:]
        x_prev = x
        (x, h) = self.mixer(self.norm(x.flatten(2)), H, W)
        x = (1 - alpha[1]) * x_prev + alpha[1] * x

        # DWConv2
        x = (1 - alpha[2]) * x + alpha[2] * self.dwconv2(x)

        # FFN
        x = (1 - alpha[3]) * x + alpha[3] * self.ffn(x)

        return (x, h)


class EfficientViMStage(nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        depth: int,
        mlp_ratio: float,
        downsample: bool,
        ssd_expand: float,
        A_init_range: tuple[int, int],
        state_dim: int,
    ) -> None:
        super().__init__()
        if downsample is True:
            self.downsample = PatchMerging(in_dim, out_dim, ratio=4.0)
        else:
            self.downsample = nn.Identity()

        self.depth = depth
        self.blocks = nn.ModuleList(
            [
                EfficientViMBlock(
                    dim=out_dim,
                    mlp_ratio=mlp_ratio,
                    ssd_expand=ssd_expand,
                    A_init_range=A_init_range,
                    state_dim=state_dim,
                )
                for _ in range(depth)
            ]
        )

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        x = self.downsample(x)
        for blk in self.blocks:
            (x, h) = blk(x)

        return (x, h)


class EfficientViM(DetectorBackbone):
    default_size = (256, 256)
    block_group_regex = r"body\.stage(\d+)\.blocks\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        mlp_ratio = 4.0
        ssd_expand = 1.0
        embed_dim: list[int] = self.config["embed_dim"]
        depths: list[int] = self.config["depths"]
        state_dim: list[int] = self.config["state_dim"]

        self.embed_dim = embed_dim

        self.stem = Stem(self.input_channels, embed_dim[0])

        num_layers = len(depths)
        prev_dim = embed_dim[0]
        return_channels: list[int] = []
        self.body = nn.ModuleList()
        for i in range(num_layers):
            self.body.add_module(
                f"stage{i+1}",
                EfficientViMStage(
                    in_dim=prev_dim,
                    out_dim=embed_dim[i],
                    depth=depths[i],
                    mlp_ratio=mlp_ratio,
                    downsample=i > 0,
                    ssd_expand=ssd_expand,
                    A_init_range=(1, 16),
                    state_dim=state_dim[i],
                ),
            )
            return_channels.append(embed_dim[i])
            prev_dim = embed_dim[i]

        # Weights for multi-stage hidden-state Fusion
        self.weights = nn.Parameter(torch.ones(4))
        self.state_norms = nn.ModuleList(
            [
                LayerNorm1d(embed_dim[0]),
                LayerNorm1d(embed_dim[1]),
                LayerNorm1d(embed_dim[2]),
            ]
        )
        self.norm = LayerNorm2d(embed_dim[2])

        self.return_stages = self.return_stages[: len(depths)]
        self.return_channels = return_channels
        self.embedding_size = embed_dim[2]
        self.state_classifiers = nn.ModuleList(
            [
                self.create_classifier(embed_dim[0]),
                self.create_classifier(embed_dim[1]),
                self.create_classifier(embed_dim[2]),
            ]
        )
        self.classifier = self.create_classifier(embed_dim[2])

        self.max_stride = 64

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, (nn.BatchNorm2d, LayerNorm2d, LayerNorm1d)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

    def reset_classifier(self, num_classes: int) -> None:
        self.num_classes = num_classes
        self.state_classifiers = nn.ModuleList(
            [
                self.create_classifier(self.embed_dim[0]),
                self.create_classifier(self.embed_dim[1]),
                self.create_classifier(self.embed_dim[2]),
            ]
        )
        self.classifier = self.create_classifier(self.embed_dim[2])

    def transform_to_backbone(self) -> None:
        self.register_parameter("weights", None)
        self.state_norms = nn.ModuleList(
            [
                nn.Identity(),
                nn.Identity(),
                nn.Identity(),
            ]
        )
        self.norm = nn.Identity()
        self.state_classifiers = nn.ModuleList(
            [
                nn.Identity(),
                nn.Identity(),
                nn.Identity(),
            ]
        )
        self.classifier = nn.Identity()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)
        out = {}
        for name, module in self.body.named_children():
            (x, _) = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

        for idx, module in enumerate(self.norm.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> tuple[torch.Tensor, list[torch.Tensor]]:
        x = self.stem(x)
        hs = []
        for stage in self.body:
            (x, h) = stage(x)
            hs.append(h)

        return (x, hs)

    def embedding(self, x: torch.Tensor) -> tuple[torch.Tensor, list[torch.Tensor]]:
        (x, hs) = self.forward_features(x)
        x = self.norm(x)
        x = F.adaptive_avg_pool2d(x, 1).flatten(1)

        return (x, hs)

    # pylint: disable=arguments-differ
    def classify(self, x: torch.Tensor, hs: list[torch.Tensor]) -> torch.Tensor:  # type: ignore[override]
        weights = F.softmax(self.weights, dim=-1)
        z = torch.zeros((x.size(0), self.num_classes), device=x.device)
        for i, (norm, classifier) in enumerate(zip(self.state_norms, self.state_classifiers)):
            h = norm(hs[i])
            h = F.adaptive_avg_pool1d(h, 1).flatten(1)  # pylint: disable=not-callable
            z = z + weights[i] * classifier(h)

        z = z + weights[3] * self.classifier(x)

        return z

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (x, hs) = self.embedding(x)
        return self.classify(x, hs)


registry.register_model_config(
    "efficientvim_m1",
    EfficientViM,
    config={"embed_dim": [128, 192, 320], "depths": [2, 2, 2], "state_dim": [49, 25, 9]},
)
registry.register_model_config(
    "efficientvim_m2",
    EfficientViM,
    config={"embed_dim": [128, 256, 512], "depths": [2, 2, 2], "state_dim": [49, 25, 9]},
)
registry.register_model_config(
    "efficientvim_m3",
    EfficientViM,
    config={"embed_dim": [224, 320, 512], "depths": [2, 2, 2], "state_dim": [49, 25, 9]},
)
registry.register_model_config(
    "efficientvim_m4",
    EfficientViM,
    config={"embed_dim": [224, 320, 512], "depths": [3, 4, 2], "state_dim": [64, 32, 16]},
)

registry.register_weights(
    "efficientvim_m1_il-common",
    {
        "url": "https://huggingface.co/birder-project/efficientvim_m1_il-common/resolve/main",
        "description": "EfficientViM M1 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 23.4,
                "sha256": "065c039bc07b1d59f58d1fe30a162a97f30e90d909d35d0df8966abaa26d8999",
            }
        },
        "net": {"network": "efficientvim_m1", "tag": "il-common"},
    },
)



"""
Inception-ResNet v2, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/inception_resnet_v2.py

Paper "Inception-v4, Inception-ResNet and the Impact of Residual Connections on Learning",
https://arxiv.org/abs/1602.07261
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation

from birder.net.base import DetectorBackbone


class StemBlock(nn.Module):
    def __init__(self, in_channels: int) -> None:
        super().__init__()
        self.branch0 = Conv2dNormActivation(
            in_channels, 96, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False
        )
        self.branch1 = nn.Sequential(
            Conv2dNormActivation(in_channels, 48, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
            Conv2dNormActivation(48, 64, kernel_size=(5, 5), stride=(1, 1), padding=(2, 2), bias=False),
        )
        self.branch2 = nn.Sequential(
            Conv2dNormActivation(in_channels, 64, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
            Conv2dNormActivation(64, 96, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False),
            Conv2dNormActivation(96, 96, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False),
        )
        self.branch_pool = nn.Sequential(
            nn.AvgPool2d(kernel_size=(3, 3), stride=(1, 1), padding=(1, 1)),
            Conv2dNormActivation(in_channels, 64, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        branch0 = self.branch0(x)
        branch1 = self.branch1(x)
        branch2 = self.branch2(x)
        branch_pool = self.branch_pool(x)
        x = torch.concat((branch0, branch1, branch2, branch_pool), dim=1)

        return x


class InceptionBlockA(nn.Module):
    def __init__(self, in_channels: int, scale: float) -> None:
        super().__init__()
        self.scale = scale
        self.branch_1x1 = Conv2dNormActivation(
            in_channels, 32, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False
        )
        self.branch_3x3 = nn.Sequential(
            Conv2dNormActivation(in_channels, 32, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
            Conv2dNormActivation(32, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False),
        )
        self.branch_3x3dbl = nn.Sequential(
            Conv2dNormActivation(in_channels, 32, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
            Conv2dNormActivation(32, 48, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False),
            Conv2dNormActivation(48, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False),
        )

        self.conv2d = nn.Conv2d(128, 320, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x
        branch_1x1 = self.branch_1x1(x)
        branch_3x3 = self.branch_3x3(x)
        branch_3x3dbl = self.branch_3x3dbl(x)
        x = torch.concat((branch_1x1, branch_3x3, branch_3x3dbl), dim=1)
        x = self.conv2d(x)
        x = (x * self.scale) + identity
        x = self.relu(x)

        return x


class InceptionReductionBlockA(nn.Module):
    def __init__(self, in_channels: int) -> None:
        super().__init__()
        self.branch0 = Conv2dNormActivation(
            in_channels, 384, kernel_size=(3, 3), stride=(2, 2), padding=(0, 0), bias=False
        )
        self.branch1 = nn.Sequential(
            Conv2dNormActivation(in_channels, 256, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
            Conv2dNormActivation(256, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False),
            Conv2dNormActivation(256, 384, kernel_size=(3, 3), stride=(2, 2), padding=(0, 0), bias=False),
        )
        self.branch_pool = nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(0, 0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        branch0 = self.branch0(x)
        branch1 = self.branch1(x)
        branch_pool = self.branch_pool(x)
        x = torch.concat((branch0, branch1, branch_pool), dim=1)

        return x


class InceptionBlockB(nn.Module):
    def __init__(self, in_channels: int, scale: float) -> None:
        super().__init__()
        self.scale = scale
        self.branch_1x1 = Conv2dNormActivation(
            in_channels, 192, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False
        )
        self.branch_7x7 = nn.Sequential(
            Conv2dNormActivation(in_channels, 128, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
            Conv2dNormActivation(128, 160, kernel_size=(1, 7), stride=(1, 1), padding=(0, 3), bias=False),
            Conv2dNormActivation(160, 192, kernel_size=(7, 1), stride=(1, 1), padding=(3, 0), bias=False),
        )

        self.conv2d = nn.Conv2d(384, 1088, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x
        branch_1x1 = self.branch_1x1(x)
        branch_7x7 = self.branch_7x7(x)
        x = torch.concat((branch_1x1, branch_7x7), dim=1)
        x = self.conv2d(x)
        x = (x * self.scale) + identity
        x = self.relu(x)

        return x


class InceptionReductionBlockB(nn.Module):
    def __init__(self, in_channels: int) -> None:
        super().__init__()
        self.branch0 = nn.Sequential(
            Conv2dNormActivation(in_channels, 256, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
            Conv2dNormActivation(256, 384, kernel_size=(3, 3), stride=(2, 2), padding=(0, 0), bias=False),
        )
        self.branch1 = nn.Sequential(
            Conv2dNormActivation(in_channels, 256, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
            Conv2dNormActivation(256, 288, kernel_size=(3, 3), stride=(2, 2), padding=(0, 0), bias=False),
        )
        self.branch2 = nn.Sequential(
            Conv2dNormActivation(in_channels, 256, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
            Conv2dNormActivation(256, 288, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False),
            Conv2dNormActivation(288, 320, kernel_size=(3, 3), stride=(2, 2), padding=(0, 0), bias=False),
        )
        self.branch_pool = nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(0, 0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        branch0 = self.branch0(x)
        branch1 = self.branch1(x)
        branch2 = self.branch2(x)
        branch_pool = self.branch_pool(x)
        x = torch.concat((branch0, branch1, branch2, branch_pool), dim=1)

        return x


class InceptionBlockC(nn.Module):
    def __init__(self, in_channels: int, scale: float, last_relu: bool) -> None:
        super().__init__()
        self.scale = scale
        self.last_relu = last_relu
        self.branch_1x1 = Conv2dNormActivation(
            in_channels, 192, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False
        )
        self.branch_3x3 = nn.Sequential(
            Conv2dNormActivation(in_channels, 192, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
            Conv2dNormActivation(192, 224, kernel_size=(1, 3), stride=(1, 1), padding=(0, 1), bias=False),
            Conv2dNormActivation(224, 256, kernel_size=(3, 1), stride=(1, 1), padding=(1, 0), bias=False),
        )

        self.conv2d = nn.Conv2d(448, 2080, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x
        branch_1x1 = self.branch_1x1(x)
        branch_3x3 = self.branch_3x3(x)
        x = torch.concat((branch_1x1, branch_3x3), dim=1)
        x = self.conv2d(x)
        x = (x * self.scale) + identity
        if self.last_relu is True:
            x = self.relu(x)

        return x


# pylint: disable=invalid-name
class Inception_ResNet_v2(DetectorBackbone):
    default_size = (299, 299)
    auto_register = True

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is None, "config not supported"

        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels, 32, kernel_size=(3, 3), stride=(2, 2), padding=(0, 0), bias=False
            ),
            Conv2dNormActivation(32, 32, kernel_size=(3, 3), stride=(1, 1), padding=(0, 0), bias=False),
            Conv2dNormActivation(32, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False),
            nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(0, 0)),
            Conv2dNormActivation(64, 80, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
            Conv2dNormActivation(80, 192, kernel_size=(3, 3), stride=(1, 1), padding=(0, 0), bias=False),
            nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(0, 0)),
        )

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []

        # Stage 1 (stem stage)
        stages["stage1"] = StemBlock(192)
        return_channels.append(320)

        # Stage 2
        layers = []
        for _ in range(10):
            layers.append(InceptionBlockA(320, scale=0.17))

        stages["stage2"] = nn.Sequential(*layers)
        return_channels.append(320)

        # Stage 3
        layers = []
        layers.append(InceptionReductionBlockA(320))
        for _ in range(20):
            layers.append(InceptionBlockB(1088, scale=0.1))

        stages["stage3"] = nn.Sequential(*layers)
        return_channels.append(1088)

        # Stage 4
        layers = []
        layers.append(InceptionReductionBlockB(1088))
        for _ in range(9):
            layers.append(InceptionBlockC(2080, scale=0.2, last_relu=True))

        layers.append(InceptionBlockC(2080, scale=1.0, last_relu=False))
        layers.append(Conv2dNormActivation(2080, 1536, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False))
        stages["stage4"] = nn.Sequential(*layers)
        return_channels.append(1536)

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
            nn.Dropout(p=0.2),
        )
        self.return_channels = return_channels[1:]
        self.return_stages = self.return_stages[1:]
        self.embedding_size = 1536
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)



"""
Dual Path Networks, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/dpn.py

Paper "Dual Path Networks", https://arxiv.org/abs/1707.01629
"""

# Reference license: Apache-2.0

from typing import Any
from typing import Literal
from typing import Optional
from typing import overload

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation

from birder.model_registry import registry
from birder.net.base import BaseNet


class NormActConv2d(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
        groups: int = 1,
        bias: bool = True,
    ) -> None:
        super().__init__()
        self.norm = nn.BatchNorm2d(in_channels)
        self.act = nn.ReLU()
        self.conv = nn.Conv2d(
            in_channels, out_channels, kernel_size=kernel_size, stride=stride, padding=padding, groups=groups, bias=bias
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.norm(x)
        x = self.act(x)
        x = self.conv(x)

        return x


class DualPathBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        num_1x1_a: int,
        num_3x3_b: int,
        num_1x1_c: int,
        increase: int,
        groups: int,
        block_type: Literal["proj", "down", "normal"],
    ) -> None:
        super().__init__()
        self.num_1x1_c = num_1x1_c
        if block_type == "proj":
            key_stride = 1
            has_proj = True
        elif block_type == "down":
            key_stride = 2
            has_proj = True
        elif block_type == "normal":
            key_stride = 1
            has_proj = False
        else:
            raise ValueError(f"Unknown block type {block_type}")

        self.c1x1_w = None
        if has_proj is True:
            self.c1x1_w = NormActConv2d(
                in_channels,
                num_1x1_c + 2 * increase,
                kernel_size=(1, 1),
                stride=(key_stride, key_stride),
                padding=(0, 0),
                bias=False,
            )

        self.c1x1_a = NormActConv2d(
            in_channels, num_1x1_a, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False
        )
        self.c3x3_b = NormActConv2d(
            num_1x1_a,
            num_3x3_b,
            kernel_size=(3, 3),
            stride=(key_stride, key_stride),
            padding=(1, 1),
            groups=groups,
            bias=False,
        )
        self.c1x1_c = NormActConv2d(
            num_3x3_b, num_1x1_c + increase, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False
        )

    # pylint: disable=protected-access

    @overload
    @torch.jit._overload_method  # type: ignore[misc]
    def forward(self, x: tuple[torch.Tensor, torch.Tensor]) -> tuple[torch.Tensor, torch.Tensor]:
        pass

    @overload
    @torch.jit._overload_method  # type: ignore[misc]
    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        pass

    # pylint: enable=protected-access

    def forward(self, x: torch.Tensor | tuple[torch.Tensor, torch.Tensor]) -> tuple[torch.Tensor, torch.Tensor]:
        if isinstance(x, tuple):
            x_in = torch.concat(x, dim=1)
        else:
            x_in = x

        if self.c1x1_w is None:
            x_s1 = x[0]
            x_s2 = x[1]
        else:
            x_s = self.c1x1_w(x_in)
            x_s1 = x_s[:, : self.num_1x1_c, :, :]
            x_s2 = x_s[:, self.num_1x1_c :, :, :]

        x_in = self.c1x1_a(x_in)
        x_in = self.c3x3_b(x_in)
        x_in = self.c1x1_c(x_in)
        out1 = x_in[:, : self.num_1x1_c, :, :]
        out2 = x_in[:, self.num_1x1_c :, :, :]

        res = x_s1 + out1
        dense = torch.concat([x_s2, out2], dim=1)

        return (res, dense)


class DPN(BaseNet):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        bw_list = [64, 128, 256, 512]
        num_init_features: int = self.config["num_init_features"]
        k_sec: list[int] = self.config["k_sec"]
        inc_sec: list[int] = self.config["inc_sec"]
        k_r: int = self.config["k_r"]
        groups: int = self.config["groups"]

        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels,
                num_init_features,
                kernel_size=(7, 7),
                stride=(2, 2),
                padding=(3, 3),
                bias=False,
            ),
            nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
        )

        layers = []
        num_features = num_init_features
        for idx in range(4):
            bw = bw_list[idx] * 4
            inc = inc_sec[idx]
            r = (k_r * bw) // (64 * 4)
            layers.append(DualPathBlock(num_features, r, r, bw, inc, groups, "proj"))
            num_features = bw + 3 * inc
            for _ in range(2, k_sec[idx] + 1):
                layers.append(DualPathBlock(num_features, r, r, bw, inc, groups, "normal"))
                num_features += inc

        self.norm_act = nn.Sequential(nn.BatchNorm2d(num_features), nn.ELU())

        self.body = nn.Sequential(*layers)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.embedding_size = num_features
        self.classifier = self.create_classifier()

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        x = self.body(x)
        x = torch.concat(x, dim=1)
        x = self.norm_act(x)

        return x

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "dpn_92",
    DPN,
    config={"num_init_features": 64, "k_sec": [3, 4, 20, 3], "inc_sec": [16, 32, 24, 128], "k_r": 96, "groups": 32},
)
registry.register_model_config(
    "dpn_98",
    DPN,
    config={"num_init_features": 96, "k_sec": [3, 6, 20, 3], "inc_sec": [16, 32, 32, 128], "k_r": 160, "groups": 40},
)
registry.register_model_config(
    "dpn_131",
    DPN,
    config={"num_init_features": 128, "k_sec": [4, 8, 28, 3], "inc_sec": [16, 32, 32, 128], "k_r": 160, "groups": 40},
)



import logging
import typing
import unittest

import torch
from torchvision.transforms import v2

from birder.data.transforms import classification
from birder.data.transforms import detection

logging.disable(logging.CRITICAL)


class TestTransforms(unittest.TestCase):
    def test_classification(self) -> None:
        # Get rgb
        for rgb_mode in typing.get_args(classification.RGBMode):
            rgb_stats = classification.get_rgb_stats(rgb_mode)
            self.assertIsInstance(rgb_stats, dict)
            self.assertIn("mean", rgb_stats)
            self.assertIn("std", rgb_stats)
            self.assertEqual(len(rgb_stats["mean"]), 3)
            self.assertEqual(len(rgb_stats["std"]), 3)

        # Get mixup / cutmix
        mixup_cutmix: v2.Transform = classification.get_mixup_cutmix(0.5, 5, True)
        self.assertIsInstance(mixup_cutmix, v2.Transform)
        self.assertEqual(len(mixup_cutmix.transforms), 3)  # identity, mixup, cutmix
        self.assertIsInstance(mixup_cutmix.transforms[0], v2.Identity)

        mixup_cutmix = classification.get_mixup_cutmix(None, 5, False)
        self.assertIsInstance(mixup_cutmix, v2.Transform)
        self.assertEqual(len(mixup_cutmix.transforms), 1)  # Only identity
        self.assertIsInstance(mixup_cutmix.transforms[0], v2.Identity)

        # Mixup module
        mixup = classification.RandomMixup(5, 0.2, 1.0)
        (samples, targets) = mixup(torch.rand((2, 3, 96, 96)), torch.tensor([0, 1], dtype=torch.int64))
        self.assertSequenceEqual(targets.size(), (2, 5))
        self.assertSequenceEqual(samples.size(), (2, 3, 96, 96))
        repr(mixup)

        # Presets
        classification.training_preset((256, 256), "birder", 0, classification.get_rgb_stats("none"))
        classification.training_preset((256, 256), "birder", 8, classification.get_rgb_stats("birder"))
        classification.training_preset((256, 256), "3aug", 3, classification.get_rgb_stats("none"))
        classification.inference_preset((256, 256), classification.get_rgb_stats("none"), 0.9)
        classification.inference_preset((256, 256), classification.get_rgb_stats("none"), 0.9, True)

    def test_detection(self) -> None:
        # Presets
        detection.training_preset((256, 256), "birder", 0, classification.get_rgb_stats("none"), False, False)
        detection.training_preset((256, 256), "birder", 5, classification.get_rgb_stats("birder"), False, True)
        detection.training_preset((256, 256), "ssd", 0, classification.get_rgb_stats("none"), True, False)
        detection.training_preset((256, 256), "multiscale", 0, classification.get_rgb_stats("none"), False, False)
        detection.InferenceTransform((256, 256), classification.get_rgb_stats("birder"), False)



"""
Barlow Twins, adapted from
https://github.com/facebookresearch/barlowtwins/blob/main/main.py

Paper "Barlow Twins: Self-Supervised Learning via Redundancy Reduction", https://arxiv.org/abs/2103.03230
"""

# Reference license: MIT

from typing import Any
from typing import Optional

import torch
import torch.distributed as dist
from torch import nn

from birder.common import training_utils
from birder.net.base import BaseNet
from birder.net.ssl.base import SSLBaseNet


def off_diagonal(x: torch.Tensor) -> torch.Tensor:
    # Return a flattened view of the off-diagonal elements of a square matrix
    (n, _) = x.size()
    # assert n == m
    return x.flatten()[:-1].view(n - 1, n + 1)[:, 1:].flatten()


class BarlowTwins(SSLBaseNet):
    def __init__(
        self,
        backbone: BaseNet,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(backbone, config=config, size=size)
        assert self.config is not None, "must set config"

        projector_sizes: list[int] = self.config["projector_sizes"]
        self.off_lambda: float = self.config["off_lambda"]

        sizes = [self.backbone.embedding_size] + projector_sizes
        layers = []
        for i in range(len(sizes) - 2):
            layers.append(nn.Linear(sizes[i], sizes[i + 1], bias=False))
            layers.append(nn.BatchNorm1d(sizes[i + 1]))
            layers.append(nn.ReLU(inplace=True))

        layers.append(nn.Linear(sizes[-2], sizes[-1], bias=False))
        self.projector = nn.Sequential(*layers)

        self.bn = nn.BatchNorm1d(sizes[-1], affine=False)

    # pylint: disable=arguments-differ
    def forward(self, x1: torch.Tensor, x2: torch.Tensor) -> torch.Tensor:  # type: ignore[override]
        world_size = training_utils.get_world_size()

        z1 = self.projector(self.backbone.embedding(x1))
        z2 = self.projector(self.backbone.embedding(x2))

        # Cross-correlation matrix
        c = self.bn(z1).T @ self.bn(z2)
        c = c / (x1.size(0) * world_size)
        if training_utils.is_dist_available_and_initialized() is True:
            dist.nn.all_reduce(c)  # https://github.com/pytorch/pytorch/issues/58005#issuecomment-1778029156

        on_diag = torch.diagonal(c).add_(-1).pow_(2).sum()
        off_diag = off_diagonal(c).pow_(2).sum()
        loss = on_diag + self.off_lambda * off_diag

        return loss



"""
EfficientNet v1, adapted from
https://github.com/pytorch/vision/blob/main/torchvision/models/efficientnet.py

Paper "EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks",
https://arxiv.org/abs/1905.11946
"""

# Reference license: BSD 3-Clause

import math
from collections import OrderedDict
from functools import partial
from typing import Any
from typing import Literal
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import SqueezeExcitation
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType
from birder.net.base import make_divisible


def adjust_channels(channels: int, width: float, min_value: Optional[int] = None) -> int:
    return make_divisible(channels * width, 8, min_value)


def adjust_depth(num_layers: int, depth: float) -> int:
    return int(math.ceil(num_layers * depth))


class MBConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        expand_ratio: float,
        stochastic_depth_prob: float,
    ) -> None:
        super().__init__()

        if stride == (1, 1) and in_channels == out_channels:
            self.use_res_connect = True
        else:
            self.use_res_connect = False

        layers = []
        activation_layer = nn.SiLU

        # Expand
        expanded_channels = adjust_channels(in_channels, expand_ratio)
        if expanded_channels != in_channels:
            layers.append(
                Conv2dNormActivation(
                    in_channels,
                    expanded_channels,
                    kernel_size=(1, 1),
                    stride=(1, 1),
                    padding=(0, 0),
                    bias=False,
                    activation_layer=activation_layer,
                )
            )

        # Depthwise
        padding = ((kernel_size[0] - 1) // 2, (kernel_size[1] - 1) // 2)
        layers.append(
            Conv2dNormActivation(
                expanded_channels,
                expanded_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                groups=expanded_channels,
                bias=False,
                activation_layer=activation_layer,
            )
        )

        # Squeeze and excitation
        squeeze_channels = max(1, in_channels // 4)
        layers.append(SqueezeExcitation(expanded_channels, squeeze_channels, activation=partial(nn.SiLU, inplace=True)))

        # Project
        layers.append(
            Conv2dNormActivation(
                expanded_channels,
                out_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
                activation_layer=None,
            )
        )

        self.block = nn.Sequential(*layers)
        self.stochastic_depth = StochasticDepth(stochastic_depth_prob, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        branch = self.block(x)
        if self.use_res_connect is True:
            return x + self.stochastic_depth(branch)

        return branch


# pylint: disable=invalid-name,too-many-locals
class EfficientNet_v1(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.stage(\d+)\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        repeats = [1, 2, 2, 3, 3, 4, 1]
        in_channels = [32, 16, 24, 40, 80, 112, 192]
        out_channels = [16, 24, 40, 80, 112, 192, 320]
        kernel_size = [(3, 3), (3, 3), (5, 5), (3, 3), (5, 5), (5, 5), (3, 3)]
        expand_ratio = [1, 6, 6, 6, 6, 6, 6]
        strides = [(1, 1), (2, 2), (2, 2), (2, 2), (1, 1), (2, 2), (1, 1)]
        width_coefficient: float = self.config["width_coefficient"]
        depth_coefficient: float = self.config["depth_coefficient"]
        dropout_rate: float = self.config["dropout_rate"]
        drop_path_rate: float = self.config.get("drop_path_rate", 0.2)

        self.dropout_rate = dropout_rate
        in_channels = [adjust_channels(ch, width_coefficient) for ch in in_channels]
        out_channels = [adjust_channels(ch, width_coefficient) for ch in out_channels]
        repeats = [adjust_depth(re, depth_coefficient) for re in repeats]

        self.stem = Conv2dNormActivation(
            self.input_channels,
            in_channels[0],
            kernel_size=(3, 3),
            stride=(2, 2),
            padding=(1, 1),
            bias=False,
            activation_layer=nn.SiLU,
        )

        layers: list[nn.Module] = []
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        stage_id = 0
        total_stage_blocks = sum(repeats)
        stage_block_id = 0
        for i, repeat in enumerate(repeats):
            for r in range(repeat):
                # Adjust stochastic depth probability based on the depth of the stage block
                sd_prob = drop_path_rate * float(stage_block_id) / total_stage_blocks

                if r > 0:
                    in_ch = out_channels[i]
                    stride = (1, 1)

                else:
                    in_ch = in_channels[i]
                    stride = strides[i]

                if stride[0] > 1 or stride[1] > 1:
                    stages[f"stage{stage_id}"] = nn.Sequential(*layers)
                    return_channels.append(in_ch)
                    layers = []
                    stage_id += 1

                layers.append(
                    MBConv(
                        in_ch,
                        out_channels[i],
                        kernel_size=kernel_size[i],
                        stride=stride,
                        expand_ratio=expand_ratio[i],
                        stochastic_depth_prob=sd_prob,
                    )
                )

                stage_block_id += 1

        stages[f"stage{stage_id}"] = nn.Sequential(*layers)
        return_channels.append(out_channels[-1])

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            Conv2dNormActivation(
                out_channels[-1],
                out_channels[-1] * 4,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
                activation_layer=nn.SiLU,
            ),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels[1:5]
        self.embedding_size = out_channels[-1] * 4
        self.classifier = self.create_classifier()

        self.stem_stride = 2
        self.stem_width = in_channels[0]
        self.encoding_size = out_channels[-1]

        # Weights initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

            elif isinstance(m, nn.Linear):
                init_range = 1.0 / math.sqrt(m.out_features)
                nn.init.uniform_(m.weight, -init_range, init_range)
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(x, mask, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token)
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def create_classifier(self, embed_dim: Optional[int] = None) -> nn.Module:
        if self.num_classes == 0:
            return nn.Identity()

        if embed_dim is None:
            embed_dim = self.embedding_size

        return nn.Sequential(
            nn.Dropout(p=self.dropout_rate, inplace=True),
            nn.Linear(embed_dim, self.num_classes),
        )


registry.register_model_config(
    "efficientnet_v1_b0",
    EfficientNet_v1,  # input_resolution = 224
    config={"width_coefficient": 1.0, "depth_coefficient": 1.0, "dropout_rate": 0.2},
)
registry.register_model_config(
    "efficientnet_v1_b1",
    EfficientNet_v1,  # input_resolution = 240
    config={"width_coefficient": 1.0, "depth_coefficient": 1.1, "dropout_rate": 0.2},
)
registry.register_model_config(
    "efficientnet_v1_b2",
    EfficientNet_v1,  # input_resolution = 260
    config={"width_coefficient": 1.1, "depth_coefficient": 1.2, "dropout_rate": 0.3},
)
registry.register_model_config(
    "efficientnet_v1_b3",
    EfficientNet_v1,  # input_resolution = 300
    config={"width_coefficient": 1.2, "depth_coefficient": 1.4, "dropout_rate": 0.3},
)
registry.register_model_config(
    "efficientnet_v1_b4",
    EfficientNet_v1,  # input_resolution = 380
    config={"width_coefficient": 1.4, "depth_coefficient": 1.8, "dropout_rate": 0.4},
)
registry.register_model_config(
    "efficientnet_v1_b5",
    EfficientNet_v1,  # input_resolution = 456
    config={"width_coefficient": 1.6, "depth_coefficient": 2.2, "dropout_rate": 0.4},
)
registry.register_model_config(
    "efficientnet_v1_b6",
    EfficientNet_v1,  # input_resolution = 528
    config={"width_coefficient": 1.8, "depth_coefficient": 2.6, "dropout_rate": 0.5},
)
registry.register_model_config(
    "efficientnet_v1_b7",
    EfficientNet_v1,  # input_resolution = 600
    config={"width_coefficient": 2.0, "depth_coefficient": 3.1, "dropout_rate": 0.5},
)

registry.register_weights(
    "efficientnet_v1_b0_il-common",
    {
        "description": "EfficientNet v1 B0 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 17.4,
                "sha256": "9b8da7fa8095cb0bec88df85f56735d8c4a41e0f68cbf5dc610665fb30a5fbbc",
            }
        },
        "net": {"network": "efficientnet_v1_b0", "tag": "il-common"},
    },
)
registry.register_weights(
    "efficientnet_v1_b1_il-common",
    {
        "description": "EfficientNet v1 B1 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 27.1,
                "sha256": "e93b5f13cd93654fbec1455a3593a963aa415c21ad28c442583af57ac43e032f",
            }
        },
        "net": {"network": "efficientnet_v1_b1", "tag": "il-common"},
    },
)
registry.register_weights(
    "efficientnet_v1_b2_il-common",
    {
        "description": "EfficientNet v1 B2 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 31.8,
                "sha256": "3a62a94364c874e9d6b8e32aeb27470d88c6b3c15d690ee623e8a59697e4cba5",
            }
        },
        "net": {"network": "efficientnet_v1_b2", "tag": "il-common"},
    },
)



"""
EfficientFormer v2, adapted from
https://github.com/snap-research/EfficientFormer/blob/main/models/efficientformer_v2.py
and
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/efficientformer_v2.py

Paper "Rethinking Vision Transformers for MobileNet Size and Speed",
https://arxiv.org/abs/2212.08059

Changes from original:
* Removed attention bias cache
"""

# Reference license: Apache-2.0 (both)

import math
from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import StochasticDepth

from birder.layers import LayerScale2d
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import interpolate_attention_bias


class Attention2d(nn.Module):
    def __init__(
        self,
        dim: int,
        key_dim: int,
        num_heads: int,
        attn_ratio: float,
        resolution: tuple[int, int],
        stride: Optional[int] = None,
    ) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.scale = key_dim**-0.5

        if stride is not None:
            resolution = (math.ceil(resolution[0] / stride), math.ceil(resolution[1] / stride))
            self.stride_conv = Conv2dNormActivation(
                dim, dim, kernel_size=(3, 3), stride=(stride, stride), padding=(1, 1), groups=dim, activation_layer=None
            )
            self.upsample = nn.Upsample(scale_factor=stride, mode="bilinear")
        else:
            self.stride_conv = nn.Identity()
            self.upsample = nn.Identity()

        self.resolution = resolution
        self.N = self.resolution[0] * self.resolution[1]
        self.dh = int(attn_ratio * key_dim) * num_heads
        kh = key_dim * self.num_heads

        self.q = Conv2dNormActivation(dim, kh, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), activation_layer=None)
        self.k = Conv2dNormActivation(dim, kh, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), activation_layer=None)
        self.v = Conv2dNormActivation(
            dim, self.dh, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), activation_layer=None
        )
        self.v_local = Conv2dNormActivation(
            self.dh, self.dh, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=self.dh, activation_layer=None
        )
        self.talking_head1 = nn.Conv2d(
            self.num_heads, self.num_heads, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)
        )
        self.talking_head2 = nn.Conv2d(
            self.num_heads, self.num_heads, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)
        )

        self.act = nn.GELU()
        self.proj = Conv2dNormActivation(
            self.dh, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), activation_layer=None
        )

        pos = torch.stack(
            torch.meshgrid(torch.arange(self.resolution[0]), torch.arange(self.resolution[1]), indexing="ij")
        ).flatten(1)
        rel_pos = (pos[..., :, None] - pos[..., None, :]).abs()
        rel_pos = (rel_pos[0] * self.resolution[1]) + rel_pos[1]
        self.attention_biases = nn.Parameter(torch.zeros(num_heads, self.N))
        self.attention_bias_idxs = nn.Buffer(torch.LongTensor(rel_pos), persistent=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        B = x.size(0)  # (B, C, H, W)
        x = self.stride_conv(x)

        q = self.q(x).reshape(B, self.num_heads, -1, self.N).permute(0, 1, 3, 2)
        k = self.k(x).reshape(B, self.num_heads, -1, self.N).permute(0, 1, 2, 3)
        v = self.v(x)
        v_local = self.v_local(v)
        v = v.reshape(B, self.num_heads, -1, self.N).permute(0, 1, 3, 2)

        attn = (q @ k) * self.scale
        attn = attn + self.attention_biases[:, self.attention_bias_idxs]
        attn = self.talking_head1(attn)
        attn = attn.softmax(dim=-1)
        attn = self.talking_head2(attn)

        x = (attn @ v).transpose(2, 3)
        x = x.reshape(B, self.dh, self.resolution[0], self.resolution[1]) + v_local
        x = self.upsample(x)

        x = self.act(x)
        x = self.proj(x)

        return x


class LocalGlobalQuery(nn.Module):
    def __init__(self, in_dim: int, out_dim: int) -> None:
        super().__init__()
        self.pool = nn.AvgPool2d(kernel_size=(1, 1), stride=(2, 2), padding=(0, 0))
        self.local = nn.Conv2d(in_dim, in_dim, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), groups=in_dim)
        self.proj = Conv2dNormActivation(
            in_dim, out_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), activation_layer=None
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        local_q = self.local(x)
        pool_q = self.pool(x)
        q = local_q + pool_q
        q = self.proj(q)

        return q


class Attention2dDownsample(nn.Module):
    def __init__(
        self,
        dim: int,
        key_dim: int,
        out_dim: int,
        num_heads: int,
        attn_ratio: float,
        resolution: tuple[int, int],
    ) -> None:
        super().__init__()

        self.num_heads = num_heads
        self.scale = key_dim**-0.5
        self.resolution = resolution
        self.resolution2 = (math.ceil(self.resolution[0] / 2), math.ceil(self.resolution[1] / 2))
        self.N = self.resolution[0] * self.resolution[1]
        self.N2 = self.resolution2[0] * self.resolution2[1]  # pylint: disable=invalid-name

        self.dh = int(attn_ratio * key_dim) * num_heads
        kh = key_dim * self.num_heads

        self.q = LocalGlobalQuery(dim, kh)
        self.k = Conv2dNormActivation(dim, kh, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), activation_layer=None)
        self.v = Conv2dNormActivation(
            dim, self.dh, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), activation_layer=None
        )
        self.v_local = Conv2dNormActivation(
            self.dh, self.dh, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), groups=self.dh, activation_layer=None
        )

        self.act = nn.GELU()
        self.proj = Conv2dNormActivation(
            self.dh, out_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), activation_layer=None
        )

        self.attention_biases = nn.Parameter(torch.zeros(num_heads, self.N))
        k_pos = torch.stack(
            torch.meshgrid(torch.arange(self.resolution[0]), torch.arange(self.resolution[1]), indexing="ij")
        ).flatten(1)
        q_pos = torch.stack(
            torch.meshgrid(
                torch.arange(0, self.resolution[0], step=2), torch.arange(0, self.resolution[1], step=2), indexing="ij"
            )
        ).flatten(1)
        rel_pos = (q_pos[..., :, None] - k_pos[..., None, :]).abs()
        rel_pos = (rel_pos[0] * self.resolution[1]) + rel_pos[1]
        self.attention_bias_idxs = nn.Buffer(torch.LongTensor(rel_pos), persistent=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        B = x.size(0)  # (B, C, H, W)

        q = self.q(x).reshape(B, self.num_heads, -1, self.N2).permute(0, 1, 3, 2)
        k = self.k(x).reshape(B, self.num_heads, -1, self.N).permute(0, 1, 2, 3)
        v = self.v(x)
        v_local = self.v_local(v)
        v = v.reshape(B, self.num_heads, -1, self.N).permute(0, 1, 3, 2)

        attn = (q @ k) * self.scale
        attn = attn + self.attention_biases[:, self.attention_bias_idxs]
        attn = attn.softmax(dim=-1)

        x = (attn @ v).transpose(2, 3)
        x = x.reshape(B, self.dh, self.resolution2[0], self.resolution2[1]) + v_local
        x = self.act(x)
        x = self.proj(x)

        return x


class Downsample(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
        resolution: tuple[int, int],
        use_attn: bool,
    ) -> None:
        super().__init__()
        self.conv = Conv2dNormActivation(
            in_channels, out_channels, kernel_size=kernel_size, stride=stride, padding=padding, activation_layer=None
        )

        if use_attn is True:
            self.attn = Attention2dDownsample(
                dim=in_channels,
                key_dim=16,
                out_dim=out_channels,
                num_heads=8,
                attn_ratio=4,
                resolution=resolution,
            )
        else:
            self.attn = None  # type: ignore[assignment]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.conv(x)
        if self.attn is not None:
            return self.attn(x) + out

        return out  # type: ignore[unreachable]


class ConvMLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: int,
        drop: float,
    ) -> None:
        super().__init__()
        self.fc1 = Conv2dNormActivation(in_features, hidden_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.mid = Conv2dNormActivation(
            hidden_features, hidden_features, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=hidden_features
        )
        self.drop1 = nn.Dropout(drop)
        self.fc2 = Conv2dNormActivation(
            hidden_features, in_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), activation_layer=None
        )
        self.drop2 = nn.Dropout(drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.fc1(x)
        x = self.mid(x)
        x = self.drop1(x)
        x = self.fc2(x)
        x = self.drop2(x)

        return x


class EfficientFormerBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        mlp_ratio: int,
        proj_drop: float,
        drop_path: float,
        layer_scale_init_value: float,
        resolution: tuple[int, int],
        stride: Optional[int],
        use_attn: bool,
    ) -> None:
        super().__init__()
        self.resolution = resolution
        self.stride = stride
        self.use_attn = use_attn

        if use_attn is True:
            self.token_mixer = Attention2d(
                dim,
                key_dim=32,
                num_heads=8,
                attn_ratio=4.0,
                resolution=resolution,
                stride=stride,
            )
            self.ls1 = LayerScale2d(dim, layer_scale_init_value)
            self.drop_path1 = StochasticDepth(drop_path, mode="row")
        else:
            self.token_mixer = None  # type: ignore[assignment]
            self.ls1 = None  # type: ignore[assignment]
            self.drop_path1 = None

        self.mlp = ConvMLP(
            in_features=dim,
            hidden_features=int(dim * mlp_ratio),
            drop=proj_drop,
        )
        self.ls2 = LayerScale2d(dim, layer_scale_init_value)
        self.drop_path2 = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.token_mixer is not None:
            x = x + self.drop_path1(self.ls1(self.token_mixer(x)))

        x = x + self.drop_path2(self.ls2(self.mlp(x)))

        return x


class EfficientFormerStage(nn.Module):
    def __init__(
        self,
        dim: int,
        dim_out: int,
        depth: int,
        resolution: tuple[int, int],
        downsample: bool,
        block_stride: Optional[int],
        downsample_use_attn: bool,
        block_use_attn: bool,
        num_vit: int,
        mlp_ratios: list[int],
        proj_drop: float,
        drop_path: list[float],
        layer_scale_init_value: float,
    ) -> None:
        super().__init__()
        self.downsample = downsample
        if downsample is True:
            self.downsample_block = Downsample(
                dim,
                dim_out,
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                resolution=resolution,
                use_attn=downsample_use_attn,
            )
            dim = dim_out
            resolution = (math.ceil(resolution[0] / 2), math.ceil(resolution[1] / 2))
        else:
            assert dim == dim_out
            self.downsample_block = nn.Identity()

        blocks = []
        for block_idx in range(depth):
            remain_idx = depth - num_vit - 1
            blocks.append(
                EfficientFormerBlock(
                    dim,
                    mlp_ratio=mlp_ratios[block_idx],
                    proj_drop=proj_drop,
                    drop_path=drop_path[block_idx],
                    layer_scale_init_value=layer_scale_init_value,
                    resolution=resolution,
                    stride=block_stride,
                    use_attn=block_use_attn and block_idx > remain_idx,
                )
            )

        self.blocks = nn.Sequential(*blocks)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample_block(x)
        x = self.blocks(x)

        return x


# pylint: disable=invalid-name
class EfficientFormer_v2(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        layer_scale_init_value = 1e-5
        embed_dims: tuple[int, int, int, int] = self.config["embed_dims"]
        depths: tuple[int, int, int, int] = self.config["depths"]
        drop_path_rate: float = self.config["drop_path_rate"]
        num_vit: int = self.config["num_vit"]
        mlp_ratios: list[list[int]] = self.config["mlp_ratios"]

        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels,
                embed_dims[0] // 2,
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                activation_layer=nn.GELU,
                inplace=None,
            ),
            Conv2dNormActivation(
                embed_dims[0] // 2,
                embed_dims[0],
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                activation_layer=nn.GELU,
                inplace=None,
            ),
        )

        prev_dim = embed_dims[0]
        stride = 4
        num_stages = len(depths)
        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]
        downsample = (False,) + (True,) * (num_stages - 1)

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_stages):
            curr_resolution = (math.ceil(self.size[0] / stride), math.ceil(self.size[1] / stride))
            stages[f"stage{i+1}"] = EfficientFormerStage(
                prev_dim,
                embed_dims[i],
                depth=depths[i],
                resolution=curr_resolution,
                downsample=downsample[i],
                block_stride=2 if i == 2 else None,
                downsample_use_attn=i >= 3,
                block_use_attn=i >= 2,
                num_vit=num_vit,
                mlp_ratios=mlp_ratios[i],
                proj_drop=0.0,
                drop_path=dpr[i],
                layer_scale_init_value=layer_scale_init_value,
            )
            return_channels.append(embed_dims[i])
            prev_dim = embed_dims[i]
            if downsample[i] is True:
                stride *= 2

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.BatchNorm2d(embed_dims[-1]),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = embed_dims[-1]
        self.dist_classifier = self.create_classifier()
        self.classifier = self.create_classifier()
        self.distillation_output = False

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)

    def reset_classifier(self, num_classes: int) -> None:
        self.num_classes = num_classes
        self.dist_classifier = self.create_classifier()
        self.classifier = self.create_classifier()

    def freeze(self, freeze_classifier: bool = True, unfreeze_features: bool = False) -> None:
        for param in self.parameters():
            param.requires_grad = False

        if freeze_classifier is False:
            for param in self.classifier.parameters():
                param.requires_grad = True

            for param in self.dist_classifier.parameters():
                param.requires_grad = True

        if unfreeze_features is True:
            for param in self.features.parameters():
                param.requires_grad = True

    def transform_to_backbone(self) -> None:
        self.features = nn.Identity()
        self.classifier = nn.Identity()
        self.dist_classifier = nn.Identity()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def set_distillation_output(self, enable: bool = True) -> None:
        self.distillation_output = enable

    def classify(self, x: torch.Tensor) -> torch.Tensor:
        x_cls = self.classifier(x)
        x_dist = self.dist_classifier(x)

        if self.training is True and self.distillation_output is True:
            x = torch.stack([x_cls, x_dist], dim=1)
        else:
            # Classifier "token" as an average of both tokens (during normal training or inference)
            x = (x_cls + x_dist) / 2

        return x

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        old_size = self.size
        super().adjust_size(new_size)

        old_base = (old_size[0] // 4, old_size[1] // 4)
        new_base = (new_size[0] // 4, new_size[1] // 4)
        for stage in self.body.modules():
            if isinstance(stage, EfficientFormerStage):
                if stage.downsample is True:
                    if stage.downsample_block.attn is not None:
                        attn = stage.downsample_block.attn
                        attn.resolution = new_base
                        attn.resolution2 = (math.ceil(new_base[0] / 2), math.ceil(new_base[1] / 2))
                        attn.N = attn.resolution[0] * attn.resolution[1]
                        attn.N2 = attn.resolution2[0] * attn.resolution2[1]

                        # Interpolate attention_biases
                        attn.attention_biases = nn.Parameter(
                            interpolate_attention_bias(attn.attention_biases, old_base, new_base)
                        )

                        k_pos = torch.stack(
                            torch.meshgrid(
                                torch.arange(attn.resolution[0]), torch.arange(attn.resolution[1]), indexing="ij"
                            )
                        ).flatten(1)
                        q_pos = torch.stack(
                            torch.meshgrid(
                                torch.arange(0, attn.resolution[0], step=2),
                                torch.arange(0, attn.resolution[1], step=2),
                                indexing="ij",
                            )
                        ).flatten(1)
                        rel_pos = (q_pos[..., :, None] - k_pos[..., None, :]).abs()
                        rel_pos = (rel_pos[0] * attn.resolution[1]) + rel_pos[1]
                        attn.attention_bias_idxs = nn.Buffer(torch.LongTensor(rel_pos), persistent=False)

                    old_base = (old_base[0] // 2, old_base[1] // 2)
                    new_base = (new_base[0] // 2, new_base[1] // 2)

                for m in stage.modules():
                    if isinstance(m, EfficientFormerBlock):
                        c_old_base = old_base
                        c_new_base = new_base
                        if m.token_mixer is not None and m.use_attn is True and m.stride is not None:
                            c_old_base = (math.ceil(old_base[0] / m.stride), math.ceil(old_base[1] / m.stride))
                            c_new_base = (math.ceil(new_base[0] / m.stride), math.ceil(new_base[1] / m.stride))

                        if m.token_mixer is not None:
                            m.token_mixer.resolution = c_new_base
                            m.token_mixer.N = m.token_mixer.resolution[0] * m.token_mixer.resolution[1]

                            m.token_mixer.attention_biases = nn.Parameter(
                                interpolate_attention_bias(m.token_mixer.attention_biases, c_old_base, c_new_base)
                            )

                            pos = torch.stack(
                                torch.meshgrid(torch.arange(c_new_base[0]), torch.arange(c_new_base[1]), indexing="ij")
                            ).flatten(1)
                            rel_pos = (pos[..., :, None] - pos[..., None, :]).abs()
                            rel_pos = (rel_pos[0] * c_new_base[1]) + rel_pos[1]
                            m.token_mixer.attention_bias_idxs = nn.Buffer(torch.LongTensor(rel_pos), persistent=False)


registry.register_model_config(
    "efficientformer_v2_s0",
    EfficientFormer_v2,
    config={
        "embed_dims": (32, 48, 96, 176),
        "depths": (2, 2, 6, 4),
        "drop_path_rate": 0.0,
        "num_vit": 2,
        "mlp_ratios": [
            [4, 4],
            [4, 4],
            [4, 3, 3, 3, 4, 4],
            [4, 3, 3, 4],
        ],
    },
)
registry.register_model_config(
    "efficientformer_v2_s1",
    EfficientFormer_v2,
    config={
        "embed_dims": (32, 48, 120, 224),
        "depths": (3, 3, 9, 6),
        "drop_path_rate": 0.0,
        "num_vit": 2,
        "mlp_ratios": [
            [4, 4, 4],
            [4, 4, 4],
            [4, 4, 3, 3, 3, 3, 4, 4, 4],
            [4, 4, 3, 3, 4, 4],
        ],
    },
)
registry.register_model_config(
    "efficientformer_v2_s2",
    EfficientFormer_v2,
    config={
        "embed_dims": (32, 64, 144, 288),
        "depths": (4, 4, 12, 8),
        "drop_path_rate": 0.02,
        "num_vit": 4,
        "mlp_ratios": [
            [4, 4, 4, 4],
            [4, 4, 4, 4],
            [4, 4, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4],
            [4, 4, 3, 3, 3, 3, 4, 4],
        ],
    },
)
registry.register_model_config(
    "efficientformer_v2_l",
    EfficientFormer_v2,
    config={
        "embed_dims": (40, 80, 192, 384),
        "depths": (5, 5, 15, 10),
        "drop_path_rate": 0.1,
        "num_vit": 6,
        "mlp_ratios": [
            [4, 4, 4, 4, 4],
            [4, 4, 4, 4, 4],
            [4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4],
            [4, 4, 4, 3, 3, 3, 3, 4, 4, 4],
        ],
    },
)

registry.register_weights(
    "efficientformer_v2_s0_il-common",
    {
        "description": "EfficientFormer v2 S0 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 13.2,
                "sha256": "b5ba923d351d45a04686b5bda037438719e0f442a41a34207a7f19737a8edb45",
            }
        },
        "net": {"network": "efficientformer_v2_s0", "tag": "il-common"},
    },
)
registry.register_weights(
    "efficientformer_v2_s1_il-common",
    {
        "description": "EfficientFormer v2 S1 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 22.9,
                "sha256": "6b7ce6bbf5aa83e222cd16d8f07e749cdbb703fd383f99e88362ec8401d81401",
            }
        },
        "net": {"network": "efficientformer_v2_s1", "tag": "il-common"},
    },
)



import pkgutil

import birder.scripts


def list_scripts() -> list[str]:
    scripts = []
    for _, name, is_pkg in pkgutil.iter_modules(birder.scripts.__path__):
        if name.startswith("_") is True:  # Skip private modules
            continue

        if is_pkg is False:
            scripts.append(name)

    return scripts


def main() -> None:
    print("Available birder scripts:")
    for script in list_scripts():
        print(f"    birder.scripts.{script}")


if __name__ == "__main__":
    main()



from birder.net.alexnet import AlexNet
from birder.net.biformer import BiFormer
from birder.net.cait import CaiT
from birder.net.cas_vit import CAS_ViT
from birder.net.coat import CoaT
from birder.net.conv2former import Conv2Former
from birder.net.convmixer import ConvMixer
from birder.net.convnext_v1 import ConvNeXt_v1
from birder.net.convnext_v2 import ConvNeXt_v2
from birder.net.crossformer import CrossFormer
from birder.net.crossvit import CrossViT
from birder.net.cspnet import CSPNet
from birder.net.cswin_transformer import CSWin_Transformer
from birder.net.darknet import Darknet
from birder.net.davit import DaViT
from birder.net.deit import DeiT
from birder.net.deit3 import DeiT3
from birder.net.densenet import DenseNet
from birder.net.dpn import DPN
from birder.net.edgenext import EdgeNeXt
from birder.net.edgevit import EdgeViT
from birder.net.efficientformer_v1 import EfficientFormer_v1
from birder.net.efficientformer_v2 import EfficientFormer_v2
from birder.net.efficientnet_lite import EfficientNet_Lite
from birder.net.efficientnet_v1 import EfficientNet_v1
from birder.net.efficientnet_v2 import EfficientNet_v2
from birder.net.efficientvim import EfficientViM
from birder.net.efficientvit_mit import EfficientViT_MIT
from birder.net.efficientvit_msft import EfficientViT_MSFT
from birder.net.fasternet import FasterNet
from birder.net.fastvit import FastViT
from birder.net.flexivit import FlexiViT
from birder.net.focalnet import FocalNet
from birder.net.ghostnet_v1 import GhostNet_v1
from birder.net.ghostnet_v2 import GhostNet_v2
from birder.net.groupmixformer import GroupMixFormer
from birder.net.hgnet_v1 import HGNet_v1
from birder.net.hgnet_v2 import HGNet_v2
from birder.net.hiera import Hiera
from birder.net.hieradet import HieraDet
from birder.net.hornet import HorNet
from birder.net.iformer import iFormer
from birder.net.inception_next import Inception_NeXt
from birder.net.inception_resnet_v1 import Inception_ResNet_v1
from birder.net.inception_resnet_v2 import Inception_ResNet_v2
from birder.net.inception_v3 import Inception_v3
from birder.net.inception_v4 import Inception_v4
from birder.net.levit import LeViT
from birder.net.maxvit import MaxViT
from birder.net.metaformer import MetaFormer
from birder.net.mnasnet import MNASNet
from birder.net.mobilenet_v1 import MobileNet_v1
from birder.net.mobilenet_v2 import MobileNet_v2
from birder.net.mobilenet_v3_large import MobileNet_v3_Large
from birder.net.mobilenet_v3_small import MobileNet_v3_Small
from birder.net.mobilenet_v4 import MobileNet_v4
from birder.net.mobilenet_v4_hybrid import MobileNet_v4_Hybrid
from birder.net.mobileone import MobileOne
from birder.net.mobilevit_v1 import MobileViT_v1
from birder.net.mobilevit_v2 import MobileViT_v2
from birder.net.moganet import MogaNet
from birder.net.mvit_v2 import MViT_v2
from birder.net.nextvit import NextViT
from birder.net.nfnet import NFNet
from birder.net.pit import PiT
from birder.net.pvt_v1 import PVT_v1
from birder.net.pvt_v2 import PVT_v2
from birder.net.rdnet import RDNet
from birder.net.regionvit import RegionViT
from birder.net.regnet import RegNet
from birder.net.regnet_z import RegNet_Z
from birder.net.repghost import RepGhost
from birder.net.repvgg import RepVgg
from birder.net.repvit import RepViT
from birder.net.resmlp import ResMLP
from birder.net.resnest import ResNeSt
from birder.net.resnet_v1 import ResNet_v1
from birder.net.resnet_v2 import ResNet_v2
from birder.net.resnext import ResNeXt
from birder.net.rope_deit3 import RoPE_DeiT3
from birder.net.rope_flexivit import RoPE_FlexiViT
from birder.net.rope_vit import RoPE_ViT
from birder.net.se_resnet_v1 import SE_ResNet_v1
from birder.net.se_resnet_v2 import SE_ResNet_v2
from birder.net.se_resnext import SE_ResNeXt
from birder.net.sequencer2d import Sequencer2d
from birder.net.shufflenet_v1 import ShuffleNet_v1
from birder.net.shufflenet_v2 import ShuffleNet_v2
from birder.net.simple_vit import Simple_ViT
from birder.net.smt import SMT
from birder.net.squeezenet import SqueezeNet
from birder.net.squeezenext import SqueezeNext
from birder.net.starnet import StarNet
from birder.net.swiftformer import SwiftFormer
from birder.net.swin_transformer_v1 import Swin_Transformer_v1
from birder.net.swin_transformer_v2 import Swin_Transformer_v2
from birder.net.tiny_vit import Tiny_ViT
from birder.net.transnext import TransNeXt
from birder.net.uniformer import UniFormer
from birder.net.van import VAN
from birder.net.vgg import Vgg
from birder.net.vgg_reduced import Vgg_Reduced
from birder.net.vit import ViT
from birder.net.vit_parallel import ViT_Parallel
from birder.net.vit_sam import ViT_SAM
from birder.net.wide_resnet import Wide_ResNet
from birder.net.xception import Xception
from birder.net.xcit import XCiT

__all__ = [
    "AlexNet",
    "BiFormer",
    "CaiT",
    "CAS_ViT",
    "CoaT",
    "Conv2Former",
    "ConvMixer",
    "ConvNeXt_v1",
    "ConvNeXt_v2",
    "CrossFormer",
    "CrossViT",
    "CSPNet",
    "CSWin_Transformer",
    "Darknet",
    "DaViT",
    "DeiT",
    "DeiT3",
    "DenseNet",
    "DPN",
    "EdgeNeXt",
    "EdgeViT",
    "EfficientFormer_v1",
    "EfficientFormer_v2",
    "EfficientNet_Lite",
    "EfficientNet_v1",
    "EfficientNet_v2",
    "EfficientViM",
    "EfficientViT_MIT",
    "EfficientViT_MSFT",
    "FasterNet",
    "FastViT",
    "FlexiViT",
    "FocalNet",
    "GhostNet_v1",
    "GhostNet_v2",
    "GroupMixFormer",
    "HGNet_v1",
    "HGNet_v2",
    "Hiera",
    "HieraDet",
    "HorNet",
    "iFormer",
    "Inception_NeXt",
    "Inception_ResNet_v1",
    "Inception_ResNet_v2",
    "Inception_v3",
    "Inception_v4",
    "LeViT",
    "MaxViT",
    "MetaFormer",
    "MNASNet",
    "MobileNet_v1",
    "MobileNet_v2",
    "MobileNet_v3_Large",
    "MobileNet_v3_Small",
    "MobileNet_v4",
    "MobileNet_v4_Hybrid",
    "MobileOne",
    "MobileViT_v1",
    "MobileViT_v2",
    "MogaNet",
    "MViT_v2",
    "NextViT",
    "NFNet",
    "PiT",
    "PVT_v1",
    "PVT_v2",
    "RDNet",
    "RegionViT",
    "RegNet",
    "RegNet_Z",
    "RepGhost",
    "RepVgg",
    "RepViT",
    "ResMLP",
    "ResNeSt",
    "ResNet_v1",
    "ResNet_v2",
    "ResNeXt",
    "RoPE_DeiT3",
    "RoPE_FlexiViT",
    "RoPE_ViT",
    "SE_ResNet_v1",
    "SE_ResNet_v2",
    "SE_ResNeXt",
    "Sequencer2d",
    "ShuffleNet_v1",
    "ShuffleNet_v2",
    "Simple_ViT",
    "SMT",
    "SqueezeNet",
    "SqueezeNext",
    "StarNet",
    "SwiftFormer",
    "Swin_Transformer_v1",
    "Swin_Transformer_v2",
    "Tiny_ViT",
    "TransNeXt",
    "UniFormer",
    "VAN",
    "Vgg",
    "Vgg_Reduced",
    "ViT",
    "ViT_Parallel",
    "ViT_SAM",
    "Wide_ResNet",
    "Xception",
    "XCiT",
]



"""
Paper "DeiT III: Revenge of the ViT", https://arxiv.org/abs/2204.07118
and
Paper "Vision Transformers Need Registers", https://arxiv.org/abs/2309.16588
"""

import math
from functools import partial
from typing import Any
from typing import Literal
from typing import Optional

import torch
from torch import nn

from birder.common.masking import mask_tensor
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenOmissionMixin
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenOmissionResultType
from birder.net.base import TokenRetentionResultType
from birder.net.vit import Encoder
from birder.net.vit import EncoderBlock
from birder.net.vit import PatchEmbed
from birder.net.vit import adjust_position_embedding


# pylint: disable=too-many-instance-attributes
class DeiT3(DetectorBackbone, PreTrainEncoder, MaskedTokenOmissionMixin, MaskedTokenRetentionMixin):
    block_group_regex = r"encoder\.block\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        image_size = self.size
        attention_dropout = 0.0
        dropout = 0.0
        pos_embed_special_tokens: bool = self.config.get("pos_embed_special_tokens", False)
        patch_size: int = self.config["patch_size"]
        num_layers: int = self.config["num_layers"]
        num_heads: int = self.config["num_heads"]
        hidden_dim: int = self.config["hidden_dim"]
        mlp_dim: int = self.config["mlp_dim"]
        layer_scale_init_value: Optional[float] = self.config.get("layer_scale_init_value", 1e-5)
        num_reg_tokens: int = self.config.get("num_reg_tokens", 0)
        drop_path_rate: float = self.config["drop_path_rate"]

        torch._assert(image_size[0] % patch_size == 0, "Input shape indivisible by patch size!")
        torch._assert(image_size[1] % patch_size == 0, "Input shape indivisible by patch size!")
        torch._assert(hidden_dim % num_heads == 0, "Hidden dim indivisible by num heads!")
        self.patch_size = patch_size
        self.num_layers = num_layers
        self.hidden_dim = hidden_dim
        self.num_reg_tokens = num_reg_tokens
        self.num_special_tokens = 1 + self.num_reg_tokens
        self.pos_embed_special_tokens = pos_embed_special_tokens
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, num_layers)]  # Stochastic depth decay rule

        self.conv_proj = nn.Conv2d(
            self.input_channels,
            hidden_dim,
            kernel_size=(patch_size, patch_size),
            stride=(patch_size, patch_size),
            padding=(0, 0),
            bias=True,
        )
        self.patch_embed = PatchEmbed()

        seq_length = (image_size[0] // patch_size) * (image_size[1] // patch_size)

        # Add a class token
        self.class_token = nn.Parameter(torch.zeros(1, 1, hidden_dim))
        if pos_embed_special_tokens is True:
            seq_length += 1

        # Add optional register tokens
        if self.num_reg_tokens > 0:
            self.reg_tokens = nn.Parameter(torch.zeros(1, self.num_reg_tokens, hidden_dim))
            if pos_embed_special_tokens is True:
                seq_length += self.num_reg_tokens
        else:
            self.reg_tokens = None

        # Add positional embedding
        self.pos_embedding = nn.Parameter(torch.empty(1, seq_length, hidden_dim).normal_(std=0.02))  # from BERT

        self.encoder = Encoder(
            num_layers,
            num_heads,
            hidden_dim,
            mlp_dim,
            dropout,
            attention_dropout,
            dpr,
            layer_scale_init_value=layer_scale_init_value,
        )
        self.norm = nn.LayerNorm(hidden_dim, eps=1e-6)

        self.return_stages = ["neck"]  # Actually meaningless, just for completeness
        self.return_channels = [hidden_dim]
        self.embedding_size = hidden_dim
        self.classifier = self.create_classifier()

        self.max_stride = patch_size
        self.stem_stride = patch_size
        self.stem_width = hidden_dim
        self.encoding_size = hidden_dim
        self.decoder_block = partial(
            EncoderBlock,
            16,
            mlp_dim=None,
            dropout=0,
            attention_dropout=0,
            drop_path=0,
            activation_layer=nn.GELU,
        )

        # Weight initialization
        if isinstance(self.conv_proj, nn.Conv2d):
            # Init the patchify stem
            fan_in = self.conv_proj.in_channels * self.conv_proj.kernel_size[0] * self.conv_proj.kernel_size[1]
            nn.init.trunc_normal_(self.conv_proj.weight, std=math.sqrt(1 / fan_in))
            if self.conv_proj.bias is not None:
                nn.init.zeros_(self.conv_proj.bias)

        if isinstance(self.classifier, nn.Linear):
            nn.init.zeros_(self.classifier.weight)
            nn.init.zeros_(self.classifier.bias)

    def _get_pos_embed(self, H: int, W: int) -> torch.Tensor:
        if self.dynamic_size is False:
            return self.pos_embedding

        if H == self.size[0] and W == self.size[1]:
            return self.pos_embedding

        return adjust_position_embedding(
            self.pos_embedding,
            (self.size[0] // self.patch_size, self.size[1] // self.patch_size),
            (H // self.patch_size, W // self.patch_size),
            self.num_special_tokens if self.pos_embed_special_tokens is True else 0,
            antialias=False,
        )

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        (H, W) = x.shape[-2:]

        x = self.conv_proj(x)
        x = self.patch_embed(x)

        batch_special_tokens = self.class_token.expand(x.shape[0], -1, -1)
        if self.reg_tokens is not None:
            batch_reg_tokens = self.reg_tokens.expand(x.shape[0], -1, -1)
            batch_special_tokens = torch.concat([batch_reg_tokens, batch_special_tokens], dim=1)

        if self.pos_embed_special_tokens is True:
            x = torch.concat([batch_special_tokens, x], dim=1)
            x = x + self._get_pos_embed(H, W)
        else:
            x = x + self._get_pos_embed(H, W)
            x = torch.concat([batch_special_tokens, x], dim=1)

        x = self.encoder(x)
        x = self.norm(x)

        x = x[:, self.num_special_tokens :]
        x = x.permute(0, 2, 1)
        (B, C, _) = x.size()
        x = x.reshape(B, C, H // self.patch_size, W // self.patch_size)

        return {self.return_stages[0]: x}

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.conv_proj.parameters():
            param.requires_grad = False

        self.pos_embedding.requires_grad = False

        for idx, module in enumerate(self.encoder.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_omission(
        self,
        x: torch.Tensor,
        ids_keep: Optional[torch.Tensor] = None,
        return_all_features: bool = False,
        return_keys: Literal["all", "tokens", "embedding"] = "tokens",
    ) -> TokenOmissionResultType:
        (H, W) = x.shape[-2:]

        # Reshape and permute the input tensor
        x = self.conv_proj(x)
        x = self.patch_embed(x)

        # Add pos embedding without special tokens
        pos_embedding = self._get_pos_embed(H, W)
        if self.pos_embed_special_tokens is True:
            x = x + pos_embedding[:, self.num_special_tokens :, :]
        else:
            x = x + pos_embedding

        # Mask tokens
        if ids_keep is not None:
            x = torch.gather(x, dim=1, index=ids_keep.unsqueeze(-1).repeat(1, 1, x.size(2)))

        # Append class and register tokens
        if self.pos_embed_special_tokens is True:
            cls_token = self.class_token + pos_embedding[:, self.num_reg_tokens : self.num_reg_tokens + 1, :]
        else:
            cls_token = self.class_token

        batch_class_token = cls_token.expand(x.shape[0], -1, -1)
        x = torch.concat((batch_class_token, x), dim=1)

        if self.reg_tokens is not None:
            if self.pos_embed_special_tokens is True:
                reg_tokens = self.reg_tokens + pos_embedding[:, 0 : self.num_reg_tokens, :]
            else:
                reg_tokens = self.reg_tokens

            batch_reg_tokens = reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        # Apply transformer
        if return_all_features is True:
            xs = self.encoder.forward_features(x)
            xs[-1] = self.norm(xs[-1])
            x = torch.stack(xs, dim=-1)
        else:
            x = self.encoder(x)
            x = self.norm(x)

        result: TokenOmissionResultType = {}
        if return_keys in ("all", "tokens"):
            result["tokens"] = x

        if return_keys in ("all", "embedding"):
            if return_all_features is True:
                x = x[..., -1]

            result["embedding"] = x[:, self.num_reg_tokens]

        return result

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        (H, W) = x.shape[-2:]

        x = self.conv_proj(x)
        x = mask_tensor(x, mask, mask_token=mask_token, patch_factor=self.max_stride // self.stem_stride)

        # Reshape and permute the input tensor
        x = self.patch_embed(x)

        # Expand the class token to the full batch
        batch_special_tokens = self.class_token.expand(x.shape[0], -1, -1)

        # Expand the register tokens to the full batch
        if self.reg_tokens is not None:
            batch_reg_tokens = self.reg_tokens.expand(x.shape[0], -1, -1)
            batch_special_tokens = torch.concat([batch_reg_tokens, batch_special_tokens], dim=1)

        if self.pos_embed_special_tokens is True:
            x = torch.concat([batch_special_tokens, x], dim=1)
            x = x + self._get_pos_embed(H, W)
        else:
            x = x + self._get_pos_embed(H, W)
            x = torch.concat([batch_special_tokens, x], dim=1)

        x = self.encoder(x)
        x = self.norm(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            features = x[:, self.num_special_tokens :]
            features = features.permute(0, 2, 1)
            (B, C, _) = features.size()
            features = features.reshape(B, C, H // self.patch_size, W // self.patch_size)
            result["features"] = features

        if return_keys in ("all", "embedding"):
            result["embedding"] = x[:, self.num_reg_tokens]

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        (H, W) = x.shape[-2:]

        # Reshape and permute the input tensor
        x = self.conv_proj(x)
        x = self.patch_embed(x)

        # Expand the class token to the full batch
        batch_special_tokens = self.class_token.expand(x.shape[0], -1, -1)

        # Expand the register tokens to the full batch
        if self.reg_tokens is not None:
            batch_reg_tokens = self.reg_tokens.expand(x.shape[0], -1, -1)
            batch_special_tokens = torch.concat([batch_reg_tokens, batch_special_tokens], dim=1)

        if self.pos_embed_special_tokens is True:
            x = torch.concat([batch_special_tokens, x], dim=1)
            x = x + self._get_pos_embed(H, W)
        else:
            x = x + self._get_pos_embed(H, W)
            x = torch.concat([batch_special_tokens, x], dim=1)

        x = self.encoder(x)
        x = self.norm(x)

        return x

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        x = x[:, self.num_reg_tokens]

        return x

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        assert new_size[0] % self.patch_size == 0, "Input shape indivisible by patch size!"
        assert new_size[1] % self.patch_size == 0, "Input shape indivisible by patch size!"

        old_size = self.size
        super().adjust_size(new_size)

        # Sort out sizes
        if self.pos_embed_special_tokens is True:
            num_prefix_tokens = self.num_special_tokens
        else:
            num_prefix_tokens = 0

        # Add back class tokens
        self.pos_embedding = nn.Parameter(
            adjust_position_embedding(
                self.pos_embedding,
                (old_size[0] // self.patch_size, old_size[1] // self.patch_size),
                (new_size[0] // self.patch_size, new_size[1] // self.patch_size),
                num_prefix_tokens,
            )
        )


registry.register_model_config(
    "deit3_t16",
    DeiT3,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 3,
        "hidden_dim": 192,
        "mlp_dim": 768,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "deit3_s16",
    DeiT3,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "drop_path_rate": 0.05,
    },
)
registry.register_model_config(
    "deit3_s14",
    DeiT3,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "drop_path_rate": 0.05,
    },
)
registry.register_model_config(
    "deit3_m16",
    DeiT3,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_dim": 2048,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "deit3_m14",
    DeiT3,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_dim": 2048,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "deit3_b16",
    DeiT3,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "deit3_b14",
    DeiT3,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "deit3_l16",
    DeiT3,
    config={
        "patch_size": 16,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "drop_path_rate": 0.45,
    },
)
registry.register_model_config(
    "deit3_l14",
    DeiT3,
    config={
        "patch_size": 14,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "drop_path_rate": 0.45,
    },
)
registry.register_model_config(
    "deit3_h16",
    DeiT3,
    config={
        "patch_size": 16,
        "num_layers": 32,
        "num_heads": 16,
        "hidden_dim": 1280,
        "mlp_dim": 5120,
        "drop_path_rate": 0.55,
    },
)
registry.register_model_config(
    "deit3_h14",
    DeiT3,
    config={
        "patch_size": 14,
        "num_layers": 32,
        "num_heads": 16,
        "hidden_dim": 1280,
        "mlp_dim": 5120,
        "drop_path_rate": 0.55,
    },
)

# With registers
registry.register_model_config(
    "deit3_reg4_t16",
    DeiT3,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 3,
        "hidden_dim": 192,
        "mlp_dim": 768,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "deit3_reg4_s16",
    DeiT3,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.05,
    },
)
registry.register_model_config(
    "deit3_reg4_s14",
    DeiT3,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.05,
    },
)
registry.register_model_config(
    "deit3_reg4_m16",
    DeiT3,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_dim": 2048,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "deit3_reg4_m14",
    DeiT3,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_dim": 2048,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "deit3_reg4_b16",
    DeiT3,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "deit3_reg4_b14",
    DeiT3,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "deit3_reg4_l16",
    DeiT3,
    config={
        "patch_size": 16,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.45,
    },
)
registry.register_model_config(
    "deit3_reg4_l14",
    DeiT3,
    config={
        "patch_size": 14,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.45,
    },
)
registry.register_model_config(
    "deit3_reg4_h16",
    DeiT3,
    config={
        "patch_size": 16,
        "num_layers": 32,
        "num_heads": 16,
        "hidden_dim": 1280,
        "mlp_dim": 5120,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.55,
    },
)
registry.register_model_config(
    "deit3_reg4_h14",
    DeiT3,
    config={
        "patch_size": 14,
        "num_layers": 32,
        "num_heads": 16,
        "hidden_dim": 1280,
        "mlp_dim": 5120,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.55,
    },
)

registry.register_weights(
    "deit3_t16_il-common",
    {
        "description": "DeiT3 tiny model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 21.5,
                "sha256": "6cd9749a9522f8ff61088e38702553fb1c4d2547b417c499652e3bfa6a81e77a",
            }
        },
        "net": {"network": "deit3_t16", "tag": "il-common"},
    },
)
registry.register_weights(
    "deit3_reg4_t16_il-common",
    {
        "description": "DeiT3 reg4 tiny model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 21.5,
                "sha256": "6806a5ae7d45f1c84b25e9869a9cbc7de94368fe9573dc3777acf2da8c83dc4e",
            }
        },
        "net": {"network": "deit3_reg4_t16", "tag": "il-common"},
    },
)
